package lsedit;

import java.awt.*;

import java.util.*;

import java.io.*;



public class EntityClass extends LandscapeClassObject {



    public static final String ENTITY_BASE_CLASS_ID	= "$ENTITY";



    public static final int ENTITY_STYLE_3DBOX		= 0;

    public static final int ENTITY_STYLE_2DBOX		= 1;

    public static final int ENTITY_STYLE_FILE		= 2;

    public static final int ENTITY_STYLE_DISK		= 3;

    public static final int ENTITY_STYLE_FOLDER		= 4;

    public static final int ENTITY_STYLE_SOURCEOBJ	= 5;

    public static final int ENTITY_STYLE_CLASS		= 6;

    public static final int ENTITY_STYLE_GROUP		= 7;

    public static final int ENTITY_STYLE_LABELLED_GROUP	= 8;



    public static final String[] styleName =

	{ "3D Box", "2D Box", "File", "Disk", 

	  "Folder", "Source Object", "Class", "Group", "Labelled Group"

        };



    protected double avg;



    // ----------------

    // Exported methods

    // ----------------



    // Constructor 

    public EntityClass(String id, int nid, EntityClass pec, Diagram dg) {

	this.id = id;

	this.setNid(nid); 

	this.label = id; 

	this.dg = dg;



	style = ENTITY_STYLE_3DBOX;



	addParentClass(pec);



        colour = Color.blue;		// Fill colour

        labelColour = Color.cyan;	// Label colour

    }



    public boolean hasId(String id) {

        return this.id.equals(id);

    }



    public void writeAttributes(PrintStream ps) throws IOException {

	if (attributes.size() > 0) {

	    ps.print(qt(id) + " {\n");

	    super.writeAttributes(ps);

	    ps.print("}\n\n"); 

	}

    }



    public EntityInstance newEntity(String id, Diagram dg) {

	EntityInstance e;



	switch(style)

	{

	case ENTITY_STYLE_3DBOX:

	    e = new Box3D(this, id, dg);

	    break;



	case ENTITY_STYLE_2DBOX:

	    e = new Box2D(this, id, dg);

	    break;



	case ENTITY_STYLE_FILE:

	    e = new FileEntity(this, id, dg);

	    break;



	case ENTITY_STYLE_DISK:

	    e = new DiskEntity(this, id, dg);

	    break;



	case ENTITY_STYLE_FOLDER:

	    e = new FolderEntity(this, id, dg);

	    break;



	case ENTITY_STYLE_SOURCEOBJ:

	    e = new SourceObjectEntity(this, id, dg);

	    break;



	case ENTITY_STYLE_CLASS:

	    e = new ClassEntity(this, id, dg);

	    break;



	case ENTITY_STYLE_GROUP:

	    e = new GroupEntity(this, id, dg);

	    break;



	case ENTITY_STYLE_LABELLED_GROUP:

	    e = new LabelledGroupEntity(this, id, dg);

	    break;



	default:

	    System.out.println("Internal error: Unknown EntityClass style " + style);

	    return null;

	}



	makeInstanceOfUs(e);



	return e;

    }



    // Accessor functions



    public String addParentClass(EntityClass inherits) {

	return addParentClass(inherits, ENTITY_BASE_CLASS_ID);

    }



    public void setAverage(double average) {

	avg = average;

    }



    public double getAverage() {

	return avg;

    }

}

