package lsedit;

import java.awt.*;

import java.util.*;

import java.io.*; 



// EntityInstance class

// 

// Provides most of the data and methods for an entity.

// Can't be instantiated directly. Requires a subclass to provide

// the draw and other functions based on a shape (e.g. box).

//



public abstract class EntityInstance extends LandscapeObject {



    public final static int SMALL_FONT = 0;

    public final static int REG_FONT   = 1;



    public final static int NOT_USED  = -1;

    public final static int CLOSED    = 0;

    public final static int OPEN      = 1;



    public final static String LABEL_ID =            "label";

    public final static String XPOSITION_ID =        "x";

    public final static String YPOSITION_ID =        "y";

    public final static String WIDTH_ID =            "width";

    public final static String HEIGHT_ID =           "height";

    public final static String VIEWS_ID =            "views";

    public final static String IN_ELISION_ID =       "elision";

    public final static String OUT_ELISION_ID =      "outelision";

    public final static String CLIENT_ELISION_ID =   "clientelision";

    public final static String SUPPLIER_ELISION_ID = "supplierelision";

    public final static String INTERNAL_ELISION_ID = "internalelision";

    public final static String AGG_ELISION_ID =      "aggelision";

    public final static String LINK_ID =             "navlink";

    public final static String ALT_LINK_ID =         "altnavlink";

    public final static String INPOINT_ID =          "inpoints";

    public final static String OUTPOINT_ID =         "outpoints";

    public final static String LEFTPOINT_ID =        "leftpoints";

    public final static String RIGHTPOINT_ID =       "rightpoints";

    public final static String DESC_ID =             "description";

    public final static String TITLE_ID =            "title";

    public final static String ANNOT_ID =            "annotations";

    public final static String FONTDELTA_ID =        "fontdelta";



    public static final int NO_TAB	= -1;

    public static final int PUBLIC_TAB  = 0;

    public static final int GROUP_TAB   = 1;

    public static final int PRIVATE_TAB = 2;



    protected static final int HALF_POINT = 4;



    protected static final int RSZ_NONE	= -1;

    protected static final int RSZ_NW	= 0;

    protected static final int RSZ_N	= 1;

    protected static final int RSZ_NE	= 2; 

    protected static final int RSZ_E	= 3;

    protected static final int RSZ_SE	= 4;

    protected static final int RSZ_S	= 5;

    protected static final int RSZ_SW	= 6;

    protected static final int RSZ_W	= 7;



    protected static final int DEF_CONT_CAP = 35; // Default contains capacity

    protected static final int ROOT_CONT_CAP = 1000; // root contains capacity



    protected EntityClass parentClass;



    protected int mark = 0;



    // Annotation tab constants



    protected static final double TAB_HEIGHT = 16.0; 



    protected static final int ANNOT_NONE = 'n';

    protected static final int ANNOT_NEW  = 'y';

    protected static final int ANNOT_OLD  = 'o';



    protected static final String ANNOT_NONE_STR = "n";

    protected static final String ANNOT_NEW_STR = "y";

    protected static final String ANNOT_OLD_STR = "o";



    protected static final Color PUBLIC_TAB_COLOUR = Color.green;

    protected static final Color GROUP_TAB_COLOUR  = Color.yellow;

    protected static final Color PRIVATE_TAB_COLOUR = Color.red;



    protected static final int MOUSE_NEAR_EDGE_THRESHOLD = 4;

    protected static final int SEP_THRESHOLD = 8;



    // Static font info 



    protected static final int MARGIN = 5; 

    protected static final int MIN_HEIGHT = 10;

    protected static final int MIN_WIDTH = 30;



    protected static final int CONTENTS_FLAG_DIM = 8;



    protected static final String OPEN_FAMILY = "Helvetica";

    protected static final int    OPEN_FONT_SIZE = 12;



    protected static final Font openFont = 

	new Font(OPEN_FAMILY, Font.PLAIN, OPEN_FONT_SIZE);



    protected static final String CLOSED_FAMILY = "Helvetica";

    protected static final int    CLOSED_FONT_MIN_SIZE = 6; 

    protected static final int    CLOSED_FONT_SMALL_SIZE = 10; 

    protected static final int	  NUM_FONTS = 25;

    protected static final int    START_FONT_NUM = 6;



    protected static Font closedFont[] = new Font[NUM_FONTS];



    protected static final Font smallFont = 

	new Font(CLOSED_FAMILY, Font.PLAIN, CLOSED_FONT_SMALL_SIZE);



    protected static boolean mungeIdToLabel = true;



    protected static final Color CARD_FORE_COLOUR = 

					new Color(1.0f, 1.0f, 0.85f);



    // First order attributes



    protected Vector     inElision = new Vector();

    protected Vector	 outElision = new Vector();

    protected Vector	 clientElision = new Vector();

    protected Vector	 supplierElision = new Vector();

    protected Vector	 internalElision = new Vector();



    protected EdgePoint[] topPoints, bottomPoints, leftPoints, rightPoints;



    protected double	 xLocal, yLocal, widthLocal, heightLocal;

    private double	 m_x;			// Top left x position of object relative to diagram top left (0,0)
    private double	 m_y;			// Top left y position of object relative to diagram top, left (0,0)
    private double	 m_width;		// Width of this object
    private double	 m_height;		// Height of this object

    protected int        fontDelta = 0;

    /* keep tight control over the group flags so can draw changes at a low level */

    private boolean	 m_groupFlag     = false;		// This element is a member of the group
    private boolean	 m_groupKeyFlag  = false;		// This is the active member of the group

    protected boolean    highlightFlag = false;



    protected int        tmpOpen = NOT_USED;



    protected boolean    fullLabel = true;


    protected boolean	 drewClientFlag = true;

    protected boolean	 drewSupplierFlag = true;



    protected int[] tabStatus = 

		{ ANNOT_NONE, ANNOT_NONE, ANNOT_NONE };



    protected EntityInstance visibleEntityCache = null;



    // Relation info



    // Our container (null -> $ROOT)

    protected EntityInstance containedBy;



    // Entities contained by this one

    protected Hashtable containsList;



    // List of relations for which this entity is the source 

    protected Vector srcRelList = new Vector();



    // List of relations for which this entity is the destination 

    protected Vector dstRelList = new Vector(); 



    protected int[] dstCardinals;



    // List of bends contained in us

    protected Vector bends = new Vector();





    // Temp layout storage



    Layout tempLyt;





    // Stat gathering



    protected int dstCount;





    // -----------------

    // Protected methods

    // -----------------



    protected void setFont(Graphics g, int type) {

	switch(type)

	{

	case SMALL_FONT:

	    g.setFont(smallFont);

	    return;



	default:

	    g.setFont(closedFont[START_FONT_NUM + fontDelta]);

	    return;

	}

    }



    protected void setFontDelta(int fd) {

	// Keep it in bounds



	fontDelta = Math.max(-START_FONT_NUM, 

			Math.min(fd, NUM_FONTS-START_FONT_NUM-1));

    }



    protected void doSaveForUndo(EntityInstance undoEntity) {

	// Save the things we know about, and then hand down to parent



	undoEntity.srcRelList = (Vector) srcRelList.clone(); 

	undoEntity.dstRelList = (Vector) dstRelList.clone(); 



	undoEntity.containsList = (Hashtable) containsList.clone(); 



	undoEntity.containedBy = containedBy; 



	// Now make copy of attributes; 	



	undoEntity.label = label;  

	undoEntity.description = description; 

	undoEntity.colour = colour; 

	// undoEntity.views = (views != null) ? views.clone() : null; 



	undoEntity.inElision = (Vector) inElision.clone();

	undoEntity.outElision = (Vector) outElision.clone();

	undoEntity.clientElision = (Vector) clientElision.clone();

	undoEntity.supplierElision = (Vector) supplierElision.clone();

	undoEntity.internalElision = (Vector) internalElision.clone();



	undoEntity.xLocal = xLocal;

	undoEntity.yLocal = yLocal; 

	undoEntity.widthLocal = widthLocal;

	undoEntity.heightLocal = heightLocal;



	undoEntity.setX(x());

	undoEntity.setY(y());

	undoEntity.setWidth(width()); 

	undoEntity.setHeight(height()); 



	// I/O points



	int nc = dg.numRelationClasses;



	if (topPoints != null) {

	    undoEntity.topPoints = new EdgePoint[nc];

	    for (int i = 0; i<nc; i++) {

		if (topPoints[i] != null) {

		    undoEntity.topPoints[i] = (EdgePoint) topPoints[i].clone();

		}

	    }

	}



	if (bottomPoints != null) {

	    undoEntity.bottomPoints = new EdgePoint[nc];

	    for (int i = 0; i<nc; i++) {

		if (bottomPoints[i] != null) {

		    undoEntity.bottomPoints[i] = 

			(EdgePoint) bottomPoints[i].clone();

		}

	    }

	}



	if (leftPoints != null) {

	    undoEntity.leftPoints = new EdgePoint[nc];

	    for (int i = 0; i<nc; i++) {

		if (leftPoints[i] != null) {

		    undoEntity.leftPoints[i] = 

			(EdgePoint) leftPoints[i].clone();

		}

	    }

	}



	if (rightPoints != null) {

	    undoEntity.rightPoints = new EdgePoint[nc];

	    for (int i = 0; i<nc; i++) {

		if (rightPoints[i] != null) {

		    undoEntity.rightPoints[i] = 

			(EdgePoint) rightPoints[i].clone();

		}

	    }

	}



	super.saveForUndo(undoEntity); 

    }



    protected void saveForUndo(EntityInstance undoEntity) {

        // Handle children



        Enumeration en = getChildren();



        while (en.hasMoreElements()) {

            EntityInstance e = (EntityInstance) en.nextElement();



            e.saveForUndo();

        }



	doSaveForUndo(undoEntity); 

    }



    protected void undo(EntityInstance undoEntity) {

        // Handle children



        Enumeration en = getChildren();



        while (en.hasMoreElements()) {

            EntityInstance e = (EntityInstance) en.nextElement();



            e.undo();

        }



        // Undo the things we know about, and then hand down to parent



        srcRelList = undoEntity.srcRelList; 

        dstRelList = undoEntity.dstRelList; 



        containsList = undoEntity.containsList; 



        containedBy = undoEntity.containedBy;



        // Copy back the attributes 



        label = undoEntity.label; 

        description = undoEntity.description; 

        colour = undoEntity.colour; 

        // undoEntity.views = (views != null) ? views.clone() : null;



        inElision = undoEntity.inElision; 

	outElision = undoEntity.outElision;

	clientElision = undoEntity.clientElision;

	supplierElision = undoEntity.supplierElision;

	internalElision = undoEntity.internalElision;



        xLocal = undoEntity.xLocal; 

        yLocal = undoEntity.yLocal;

        widthLocal = undoEntity.widthLocal;

        heightLocal = undoEntity.heightLocal;



        setX(undoEntity.x());

        setY(undoEntity.y());

        setWidth(undoEntity.width());

        setHeight(undoEntity.height());

	topPoints    = undoEntity.topPoints;
	bottomPoints = undoEntity.bottomPoints;
	leftPoints   = undoEntity.leftPoints;
	rightPoints  = undoEntity.rightPoints;

        super.undo(undoEntity);

    }



    protected double tabHeight() {

	return height()/3-2;

    }



    protected void handleTab(Graphics g, int tab) {



	if (tabStatus[tab] == ANNOT_NONE)

	    return;



        Color c;



        switch(tab)

        {

        case PUBLIC_TAB:

            c = PUBLIC_TAB_COLOUR;

            break;



        case GROUP_TAB:

            c = GROUP_TAB_COLOUR;

            break;



        default: 

            c = PRIVATE_TAB_COLOUR;

            break;

        }



	drawTab(g, c, tab, (tabStatus[tab] == ANNOT_NEW));

    }



    protected void drawAnnotationTabs(Graphics g) {

        if (dg.drawAnnotTabs()) {

	    handleTab(g, PUBLIC_TAB);

	    handleTab(g, GROUP_TAB);

	    handleTab(g, PRIVATE_TAB); 

        }

    }



    protected void drawIOpoint(Graphics g, 

				RelationClass rc, EdgePoint pt, boolean sc)

    {

/*

	// Cache here if needed

	// sc = set colour flag



	if (sc) {

	    Color cc = g.getColor();



	    RelationInstance.drawBend(g, pt.x, pt.y, true);



	    g.setColor(cc);

	}

	else {

	     RelationInstance.drawBend(g, pt.x, pt.y, false);

	}

*/

    }



    public void drawTabHighlight(Graphics g, int tab) {

	drawTab(g, Color.white, tab, false);

    }



    public void undrawTabHighlight(Graphics g, int tab) {

        handleTab(g, tab);

    }



    protected void setColor(Graphics g) {

	g.setColor(colour);

    }



    public Color getLabelColour() {

	if (highlightFlag && !isOpen()) {

	    return Color.yellow;

	}

	else {

	    return labelColour;

	}

    }





    //

    // Parse an attribute record to set the appropriate edge point factors

    // supplied in an attribute record.



    protected void parsePoints(EdgePoint[] ept, Attribute attr) {

	AttributeValueItem avi = attr.avi;



	while (avi != null) {

	    RelationClass rc = dg.nameToRelationClass(avi.value);



	    if (rc != null) {

	        int nid = rc.getNid();



	        double wf = Util.parseReal(avi.next.value);

	        double hf = Util.parseReal(avi.next.next.value);



		if (ept[nid] == null) {

		    ept[nid] = new EdgePoint(this, rc, wf, hf);

		}

		else {

		    if (rc != ept[nid].getRc())

			MsgOut.println("Class nid mismatch");



	            ept[nid].wf = wf;

	            ept[nid].hf = hf;

		}



		ept[nid].isDefault = false;



	        avi = avi.nextList;

	    }

	}

    }



    static protected Hashtable est = new Hashtable();



    protected void processElision(Attribute attr, Vector list) {

	AttributeValueItem av = attr.avi;



	while(av != null) {

	    String str = (String) est.get(av.value);

	    if (str == null) {

		str = av.value;



		est.put(str, str);

	    }



	    if (!list.contains(str))

	        list.addElement(str);



	    av = av.next;

	}

    }



    protected void processRawElision(String[] sl, Vector list) {

	for (int i = 0; i<sl.length; i++) {

	    list.addElement(sl[i]);

	}

    }



    //

    // Parse the values of any first order attributes into their storage.

    // 



    protected boolean processFirstOrderAttributes(Attribute attr)

    {

	boolean hasVal = (attr.avi != null);



        if (attr.hasId(XPOSITION_ID)) {

	    if (hasVal)

                xLocal = attr.parseReal();

	    return true; 

        }

        if (attr.hasId(YPOSITION_ID)) {

	    if (hasVal) 

                yLocal = attr.parseReal();

	    return true; 

        }

        if (attr.hasId(WIDTH_ID)) {

	    if (hasVal) 

                widthLocal = attr.parseReal();

	    return true; 

        }

        if (attr.hasId(HEIGHT_ID)) {

	    if (hasVal) 

                heightLocal = attr.parseReal();

	    return true; 

        }

        if (attr.hasId(LABEL_ID)) {

	    if (hasVal) 

                label = attr.parseString();

	    return true; 

        }

	if (attr.hasId(DESC_ID)) {

	    if (hasVal) 

	        description = attr.parseString();

	    return true; 

	} 

        if (attr.hasId(TITLE_ID)) {

            if (hasVal)

                title = attr.parseString();

            return true;

        }

	if (attr.hasId(FONTDELTA_ID)) {

	    if (hasVal) 

		setFontDelta(attr.parseInt());

	    return true; 

	}

	if (attr.hasId(COLOUR_ID)) {

	    if (hasVal)

                colour = attr.parseColour();

            return true;

	}

	if (attr.hasId(LABEL_COLOUR_ID)) {

	    if (hasVal)

                labelColour = attr.parseColour();

            return true;

	}

        if (attr.hasId(IN_ELISION_ID)) {

	    if (hasVal) 

	        processElision(attr, inElision); 

	    return true; 

        }

	if (attr.hasId(OUT_ELISION_ID)) {

	    if (hasVal)

		processElision(attr, outElision);

	    return true;

	}

	if (attr.hasId(CLIENT_ELISION_ID)) {

	    if (hasVal)

		processElision(attr, clientElision);

	    return true;

	}

	if (attr.hasId(SUPPLIER_ELISION_ID)) {

	    if (hasVal)

		processElision(attr, supplierElision);

	    return true;

	}

	if (attr.hasId(INTERNAL_ELISION_ID)) {

	    if (hasVal)

		processElision(attr, internalElision);

	    return true;

	}

	if (attr.hasId(INPOINT_ID)) {

	    if (hasVal) {

		if (topPoints == null)

		    topPoints = new EdgePoint[dg.numRelationClasses];

	        parsePoints(topPoints, attr);

	    }

	    return true; 

	}

	if (attr.hasId(OUTPOINT_ID)) {

	    if (hasVal) {

		if (bottomPoints == null)

		    bottomPoints = new EdgePoint[dg.numRelationClasses];

	        parsePoints(bottomPoints, attr); 

	    }

	    return true; 

	}

	if (attr.hasId(LEFTPOINT_ID)) {

	    if (hasVal) {

		if (leftPoints == null)

		    leftPoints = new EdgePoint[dg.numRelationClasses];

	        parsePoints(leftPoints, attr); 

	    }

	    return true; 

	}

	if (attr.hasId(RIGHTPOINT_ID)) {

	    if (hasVal) {

		if (rightPoints == null)

		    rightPoints = new EdgePoint[dg.numRelationClasses];

	        parsePoints(rightPoints, attr); 

	    }

	    return true; 

	}

	return false; 

    }



    protected boolean hasRelation(RelationClass rc, boolean srcList) {

        Enumeration enRels = (srcList ? srcRelList.elements() : 

					dstRelList.elements());



	while(enRels.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) enRels.nextElement();



	    if (ri.parentClass.id.equals(rc.id))

		return true; 

	}



	// Now check for infered relations based on what we contain 



	Enumeration contains = getChildren();



	while (contains.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) contains.nextElement();



	    if (e.hasRelation(rc, srcList))

		return true;

	}



	return false; 

    }





    protected void doEntityDelete() {

	// Remove dstRelList relations from entities on our srcRelList.

	// Remove srcRelList relations from entities on our dstRelList. 

	// Remove entity from parent's contain list. 



	Enumeration en = srcRelList.elements(); 



	while(en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    ri.getDst().removeDstRelation(ri); 

	} 



	en = dstRelList.elements();



	while(en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    ri.getSrc().removeSrcRelation(ri);

	}



	containedBy.removeContainment(this);

    }





    public Layout getLocalContentBoundingBox() {

	if (containsList.size() == 0) {

	    MsgOut.dprintln("No children");

	    return null;

	}



	// Determine the bounding box of all contents



	Enumeration en = getChildren();



	double x1, y1, x2, y2;



	x1 = xLocal + widthLocal; 

	y1 = yLocal + heightLocal;

	x2 = 0;

	y2 = 0;



	while (en.hasMoreElements()) {

            EntityInstance e = (EntityInstance) en.nextElement();



            if (e.xLocal < x1) 

		x1 = e.xLocal;



	    if (e.yLocal < y1)

		y1 = e.yLocal;



	    if (e.xLocal + e.widthLocal > x2)

		x2 = e.xLocal + e.widthLocal;



	    if (e.yLocal + e.heightLocal > y2)

		y2 = e.yLocal + e.heightLocal;

        }



	en = bends.elements();



	int bs = Bend.SIZE/2;



	while (en.hasMoreElements()) {

	    Bend bend = (Bend) en.nextElement();



	    if (bend.xLocal - bs < x1)

		x1 = bend.xLocal - bs;

	

	    if (bend.yLocal - bs < y1)

		y1 = bend.yLocal - bs;



	    if (bend.xLocal + bs > x2) 

		x2 = bend.xLocal + bs;



	    if (bend.yLocal + bs > y2)

		y2 = bend.yLocal + bs;

	}



	return new Layout(x1, y1, x2-x1, y2-y1);

    }

   

    protected String endSet[] = { 

	".ss", ".dup", "__funcdef", "__funcdcl", "__vardef", "__vardcl",

	"__macro", "__struct", "__union", "__enum", "__type"

    }; 



    public String mungeId(String id) {

	if (mungeIdToLabel && id.length() > 0) {



	    for (int i = 0; i<endSet.length; i++) {

	        if (id.endsWith(endSet[i])) {

		    id = id.substring(0, id.length()-endSet[i].length());

		    break;

		}

	    }



	    if (id.charAt(0) != '<') {

	        int lpi = id.lastIndexOf('/');



	        if (lpi < 0) 

		    lpi = id.lastIndexOf('\\');



	        if (lpi >= 0 && id.length() != lpi+1) {

		    return id.substring(lpi+1);

	        }

	    }

	}

	return id;

    }



    // --------------

    // Static methods

    // -------------- 


    /* Allocates helvetica fonts of increasing point size */

    static public void generateFonts() {

        for (int i=0; i<NUM_FONTS; i++) {

            closedFont[i] =

                new Font(CLOSED_FAMILY, Font.PLAIN, CLOSED_FONT_MIN_SIZE+i /* Point size */);

        }

    }



    static public void setLabelMungeFlag(boolean state) {

	mungeIdToLabel = state;

    }





    // ----------------

    // Abstract methods

    // ----------------



    // draw()

    //

    // Must be provided by a subclass of EntityInstance to handle

    // drawing of actual entity shape (e.g. box).



    abstract public void saveForUndo(); 

    abstract public void undo(); 



    abstract public void draw( Graphics g);

    abstract public void drawTab(Graphics g, Color c, int tab, boolean fill);

    // abstract public void drawIOpoint(Graphics g, EdgePoint pt);



    // Calculate the intercept between the entity and the line between

    // points p1 and p2, where p1 is in the entity.



    abstract public RealPoint calculateIntercept(RealPoint p1, RealPoint p2);



    abstract public boolean isPointOver(int x, int y);

    abstract public boolean isPointOverTab(int x, int y, int tab);

    abstract public boolean isPointOverIO(EdgePoint pt, int x, int y);

    abstract public boolean containsLayout(Layout lyt); 

    abstract public boolean intersectsLayout(Layout lyt); 

    abstract public int     overResizeTab(int x, int y);



    abstract public void drawHighlight(Graphics g);

    abstract public void undrawHighlight(Graphics g);

    abstract public void drawOutline(Graphics g, Layout lyt); 



    abstract public EdgePoint getPoint(RelationClass rc, int side);

    abstract public EdgePoint getOutPoint(RelationInstance ri, int edge_mode, Layout srcLyt, Layout dstLyt);



    abstract public EdgePoint getMouseOverIO(int x, int y);

    abstract public void writeIOpoints(PrintStream ps);



    // --------------

    // Public methods

    // --------------



    protected EntityInstance() {

    }



    protected String extendedLabel(boolean wParent) {

	if (wParent) {

	    EntityInstance pe = getEnterableParent();



	    if (pe != null) {

		return pe.getLabel() + " .\n" + label;

	    }

	}



	return label;

    }



    public void setup(EntityClass parentClass, String id, Diagram dg) {

	this.dg = dg;

	this.id = id;

	this.parentClass = parentClass;



	this.label = mungeId(id); // Default



	// A default layout



        this.setX(10);

        this.setY(10);

        this.setWidth(20);

        this.setHeight(10);



	containsList = new Hashtable(id.equals(Diagram.ROOT_ID) ?

					ROOT_CONT_CAP : DEF_CONT_CAP);



	if (parentClass != null) {

	    colour = parentClass.getColor();

	    labelColour = parentClass.getLabelColor();

	}

    }

 

    public EntityInstance(EntityClass parentClass, String id, Diagram dg) {

	setup(parentClass, id, dg);

    }



    public void clearCache() {

	// Reset any caching



	visibleEntityCache = null;

	topPoints = bottomPoints = rightPoints = leftPoints = null;

        Enumeration en = getChildren();

        while (en.hasMoreElements()) {
            EntityInstance e = (EntityInstance) en.nextElement();
            e.clearCache();
        }

    }

    public void drawAll(Graphics g) {

	drewClientFlag = false;

	drewSupplierFlag = false;

	dstCount = 0;

	draw(g);

        if (!isOpen()) {

	    if (hasChildren()) {

		drawContentsFlag(g);
	    }
	} else {
	    drawChildren(g);
	}
    }



    public void drawChildren(Graphics g) {

        if (!isOpen())

	    return;



        Enumeration en = getChildren();



        while (en.hasMoreElements()) {

            EntityInstance e = (EntityInstance) en.nextElement();



            e.drawAll(g);

        }

    }



    // Destroy just the entity, leaving the children 



    public void delete() {

	double xScale = dg.getScaleX();

	double yScale = dg.getScaleY();



	// Promote our children



	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    Layout lyt = e.getLayout();



	    containedBy.addContainment(e);



	    e.setLayout(lyt); 

	}



	doEntityDelete(); 

    }



    // Destroy entity and all it contains 



    public void deleteAll(boolean delThis) {

	// deleteAll on our children



	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.deleteAll(true);

	} 



	if (delThis)

	    doEntityDelete();

    }



    public void addContainment(EntityInstance e) {

	// This entity contains the passed entity

	// Put on this' contains list



	if (containsList.put(e.id, e) != null) {

	    MsgOut.println("Entity " + id + " already contains " + e.id);

	}

	else {

	    e.containedBy = this;

	}

    }



    public boolean removeContainment(EntityInstance e) {

	if (containsList.containsKey(e.id)) {

	    containsList.remove(e.id);

	    return true;

	}

	else {

	    MsgOut.println("Entity " + id + " doesn't contain " + e.id);

	    return false;

	}

    }



    public void moveEntityContainment(EntityInstance newContainer) {

	// Transfer containment iff we need to



	if (newContainer != containedBy) {

	    containedBy.removeContainment(this);

	    newContainer.addContainment(this); 

	}

    }



    public void addRelation(RelationInstance ri, EntityInstance dst) {

	// Should check to see if relation is already present



	addSrcRelation(ri); 

	dst.addDestRelation(ri); 

    }



    public RelationInstance getRelation(RelationClass rc, EntityInstance dst) {

	// Return relation instance if present



	Enumeration en = srcRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    if (ri.getRelationClass() == rc && ri.getDst() == dst)

		return ri;

	}



	return null;

    }



    public void addSrcRelation(RelationInstance ri) {

	srcRelList.addElement(ri); 

    }



    public void removeSrcRelation(RelationInstance ri) {

	srcRelList.removeElement(ri); 

    }



    public void addDestRelation(RelationInstance ri) {

	dstRelList.addElement(ri); 

    }



    public void removeDstRelation(RelationInstance ri) {

	dstRelList.removeElement(ri);

    }



    public boolean isRelationPresent(RelationClass rc, EntityInstance dst) {

	String dstId = dst.getId();



	Enumeration en = srcRelList.elements();



	while(en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    if (ri.isSameRelation(rc, this, dst)) {

		return true; 

	    }

	}



	return false; 

    }



    public void addAttribute(Attribute attr) {

	// Process first order attributes, and add all others 

	// to an attribute database



	if (processFirstOrderAttributes(attr))

	    return;



	super.addAttribute(attr);

    }



    public boolean addRawAttribute(String id, int val) {

	if (id.equals(FONTDELTA_ID)) {

	    setFontDelta(val);

	    return true; 

	}



	return false;

    }



    public boolean addRawAttribute(String id, double val) {

        if (id.equals(XPOSITION_ID)) {

            xLocal = val;

	    return true; 

        }

        else if (id.equals(YPOSITION_ID)) {

            yLocal = val;

	    return true; 

        }

        else if (id.equals(WIDTH_ID)) {

            widthLocal = val;

	    return true; 

        }

        else if (id.equals(HEIGHT_ID)) {

            heightLocal = val;

	    return true; 

        }



	return false;

    }



    public boolean addRawAttribute(String id, String val) {

	if (id.equals(LABEL_ID)) {

            label = val;

	    return true; 

        }

	else if (id.equals(DESC_ID)) {

	    description = val;

	    return true; 

	} 

        else if (id.equals(TITLE_ID)) {

            title = val;

            return true;

        }



	return super.addRawAttribute(id, val);

    }



    public boolean addRawAttribute(String id, int[] il) {

	return false;

    }



    public boolean addRawAttribute(String id, double[] dl) {

	if (id.equals(COLOUR_ID)) {

            colour = new Color((float) dl[0], (float) dl[1], (float) dl[2]);

            return true;

	}

	else if (id.equals(LABEL_COLOUR_ID)) {

            labelColour = 

		new Color((float) dl[0], (float) dl[1], (float) dl[2]);

            return true;

	}



	return false;

    }



    public boolean addRawAttribute(String id, String[] sl) {

	if (id.equals(IN_ELISION_ID)) {

	    processRawElision(sl, inElision); 

	    return true; 

        }

	else if (id.equals(OUT_ELISION_ID)) {

	    processRawElision(sl, outElision);

	    return true;

	}

	else if (id.equals(CLIENT_ELISION_ID)) {

	    processRawElision(sl, clientElision);

	    return true;

	}

	else if (id.equals(SUPPLIER_ELISION_ID)) {

	    processRawElision(sl, supplierElision);

	    return true;

	}

	else if (id.equals(INTERNAL_ELISION_ID)) {

	    processRawElision(sl, internalElision);

	    return true;

	}



	return super.addRawAttribute(id, sl);

    }



    public void assignAttributes(Attribute attr) {

        // Passed list of attribute assignments

  

        while (attr != null) {

            // For each attribute assignment, find the variable,

            // and assign the value.



	    if (attr.avi == null) {

		MsgOut.println("Null attribute assignment");

	    }

	    else if (!processFirstOrderAttributes(attr)) {



                Attribute attrVar = getAttribute(attr.id);



                if (attrVar == null) {

                    MsgOut.println("Missing attribute '" +

                                        attr.toString() + "'");

                }

                else {

                    if (attrVar.cloneOnAssign) {

                        attrVar = (Attribute) attrVar.clone();

                        attrVar.cloneOnAssign = false;

			replaceAttribute(attrVar);

                    }

                    attrVar.avi = attr.avi;

		}

            }

            attr = attr.next;

        }

    } 



    public void assignRelnAttributes(RelationClass rc, EntityInstance dst,

					Attribute attr)

    {	

	// Find the associated edge

	System.out.println("Asgn: " + rc.getId() + " " + id + " " + dst.getId());

    }



/*
	v6.0.3

	Stupid to rescale -- might as well recalculate 
 */


    protected void rescale(EdgePoint[] pts) {

	if (pts != null) {
	    for (int i = 0; i < pts.length; i++) {
		if (pts[i] != null) {
		    pts[i].clearRc();
    }	}   }	}



    // Global coordinate transform

    //

    // Calculate global values for layout from local with scale.

    // The absolute x and y offsets of the parent are passed.

    //

    // This also handles calling the rescale on the containing entities.

    //



    public void rescale(double xL, double yL, double wL, double hL,

			double xScale, double yScale,

			double xoffset, double yoffset)

    {

        setX(xL * xScale + xoffset);

        setY(yL * yScale + yoffset);



        setWidth(wL * xScale);

        setHeight(hL * yScale);



	// rescale bends



	Enumeration en = bends.elements();



	while (en.hasMoreElements()) {

	    Bend bend = (Bend) en.nextElement();



	    bend.x = bend.xLocal * xScale + x();

	    bend.y = bend.yLocal * yScale + y();



	    MsgOut.dprintln(id + " " + bend.x + " " + bend.y);

	}



	// rescale edge points

	rescale(topPoints);
	rescale(bottomPoints);
	rescale(leftPoints);
	rescale(rightPoints);


	en = getChildren();



	double new_xoffset = xoffset + (xL * xScale);

	double new_yoffset = yoffset + (yL * yScale); 



        while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



            e.rescale(xScale, yScale, new_xoffset, new_yoffset);

        }

    }



    public void rescale(double xScale, double yScale,

			double xoffset, double yoffset)

    {

	rescale(xLocal, yLocal, widthLocal, heightLocal, 

			xScale, yScale, xoffset, yoffset);

    }





    public void rescaleChildren(double xScale, double yScale,

			double xoffset, double yoffset) 

    {

 

	// rescale edge points

	rescale(topPoints);
	rescale(bottomPoints);
	rescale(leftPoints);
	rescale(rightPoints);



	Enumeration en = getChildren();



        while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



            e.rescale(xScale, yScale, xoffset, yoffset);

        }

    }





    protected void 

    scale(double wf, double hf, double dx, double dy, 

	  boolean doContainer, boolean doPos)

    {

	// Scale ourselves 



	if (doContainer) {

	    widthLocal *= wf;

	    heightLocal *= hf;  

	    if (doPos) {

	        xLocal = (xLocal - dx) * wf;

	        yLocal = (yLocal - dy) * hf;

	    }

	}

	else if (containsList.size() > 0) {

	    // Insure the scale of the bounding box of the contained elements

	    // will be within the container.



	    Layout bb = getLocalContentBoundingBox();



	    bb.width *= wf;

	    bb.height *= hf;



	    if (doPos) {

		bb.x = (bb.x - dx) * wf;

		bb.y = (bb.y - dy) * hf;

	    }



	    if ((bb.x + bb.width >= widthLocal) || 

		(bb.y + bb.height >= heightLocal))

	    {

		return; 	// Don't bother, it won't fit right

	    }

	}



	// ... any bends we contain ...



	Enumeration en = bends.elements();



	while(en.hasMoreElements()) {

	    Bend bend = (Bend) en.nextElement();



	    bend.xLocal = (bend.xLocal-dx) * wf;

	    bend.yLocal = (bend.yLocal-dy) * hf;

	}



	// ... and then our children



	en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.scale(wf, hf, dx, dy, true, true);

        }

    }



    public void scale(double sf, boolean doContainer, boolean doPos) {

	scale(sf, sf, 0.0, 0.0, doContainer, doPos);

    }



    public void scale(double sf, boolean doContainer) {

	scale(sf, sf, 0.0, 0.0, doContainer, false);

    }



    public void scale(double sf) {

	scale(sf, sf, 0.0, 0.0, true, false);

    }



    public void scaleX(double sf, boolean doContainer, boolean doPos) {

	scale(sf, 1.0, 0.0, 0.0, doContainer, doPos);

    }



    public void scaleX(double sf, boolean doContainer) {

	scale(sf, 1.0, 0.0, 0.0, doContainer, false);

    }



    public void scaleX(double sf) {

	scale(sf, 1.0, 0.0, 0.0, true, false);

    }



    public void scaleY(double sf, boolean doContainer, boolean doPos) {

	scale(1.0, sf, 0.0, 0.0, doContainer, doPos);

    }



    public void scaleY(double sf, boolean doContainer) {

	scale(1.0, sf, 0.0, 0.0, doContainer, false); 

    }



    public void scaleY(double sf) {

	scale(1.0, sf, 0.0, 0.0, true, false);

    }



    // Attempt to fit children to us  



    protected void doFitScale(double wf, double hf, double dx, double dy) {



	widthLocal *= wf;

	heightLocal *= hf;  

	xLocal = xLocal * wf - dx;

	yLocal = yLocal * hf - dy;



	// ... any bends we contain ...



	Enumeration en = bends.elements();



	while(en.hasMoreElements()) {

	    Bend bend = (Bend) en.nextElement();



	    bend.xLocal = (bend.xLocal-dx) * wf;

	    bend.yLocal = (bend.yLocal-dy) * hf;

	}



	// ... and then our children



	en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.doFitScale(wf, hf, 0, 0);

        }

    }



    public boolean fitTo(boolean alwaysFit) {

	double xScale = dg.getScaleX();

	double yScale = dg.getScaleY();



	MsgOut.dprintln("Fit to: " + id +  

		" (" + widthLocal + " " + heightLocal + ")");



        if ( (widthLocal == 0) || (heightLocal == 0)) {

            containsList = new Hashtable();

        }



	if (containsList.size() == 0) {

	    MsgOut.dprintln("No children");

	    return true;

	}



	// Determine the bounding box of all children



	Enumeration en = getChildren();



        double x1 = Double.MAX_VALUE;

        double x2 = Double.MIN_VALUE;

        double y1 = Double.MAX_VALUE;

        double y2 = Double.MIN_VALUE;



	while (en.hasMoreElements()) {

            EntityInstance e = (EntityInstance) en.nextElement();



            if (e.x() < x1) 

		x1 = e.x();



	    if (e.y() < y1)

		y1 = e.y();



	    if (e.x() + e.width() > x2)

		x2 = e.x() + e.width();



	    if (e.y() + e.height() > y2)

		y2 = e.y() + e.height();

        }



	en = bends.elements();



	int bs = Bend.SIZE/2;



	while (en.hasMoreElements()) {

	    Bend bend = (Bend) en.nextElement();



	    if (bend.x - bs < x1)

		x1 = bend.x - bs;

	

	    if (bend.y - bs < y1)

		y1 = bend.y - bs;



	    if (bend.x + bs > x2) 

		x2 = bend.x + bs;



	    if (bend.y + bs > y2)

		y2 = bend.y + bs;

	}

        if ((x2 - x1 ) <= 0.0001 ) {

            return false;

        }

        else

        if ((y2 - y1 ) <= 0.0001 ) {

            return false;

        }



	if (!alwaysFit) {

	    // Only continue if bb isn't completely contained in this



	    Layout lyt = new Layout(x1, y1, x2-x1, y2-y1);



	    if (containsLayout(lyt))

		return true;

	}



	MsgOut.dprintln("(" + x1 + "," + y1 + ") (" + x2 + "," + y2 + ")");



	// (x1, y1) and (x2, y2) consitute bounding box for children 



	// Translate bb to top left of parent and scale to fit





	double scaleX = (width()-width()/15)/(x2-x1);

	double scaleY = (height()-height()/15)/(y2-y1);



	double sbbw = (x2-x1)/xScale*scaleX;

	double sbbh = (y2-y1)/yScale*scaleY;



	double dx = (x1 - x())/xScale*scaleX - (widthLocal-sbbw)/2;

	double dy = (y1 - y())/yScale*scaleY - (heightLocal-sbbh)/2;



	en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.doFitScale(scaleX, scaleY, dx, dy);

        }



	MsgOut.dprintln("Fit: " + scaleX + " " + scaleY);

        return true;

    }



    public boolean verifyFit() {

	// Verify for our children

	

	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance ce = (EntityInstance) en.nextElement();



	    if( !ce.verifyFit() ) {

                containsList = new Hashtable();

	        colour = Color.black;

                labelColour = Color.white;

                setLabel("LSedit depth limit!!!");

	        MsgOut.dprintln("Hit LSedit depth limit: see [irbull]");

                break;

            }

	}

        return fitTo( false );

    }





    public void deltaFont(int delta) {

	setFontDelta(fontDelta + delta);

    }



    protected void drawTopLeftLabel(Graphics g) {

	g.setFont(smallFont);

	g.setColor(Color.black); 



	Util.drawStringClipped(g, label, x()+MARGIN, y()+MARGIN, 

					width()-MARGIN*2, height()-MARGIN*2);

    }



    public boolean doDrawLabel(Graphics g,

			        double x, double y,

				double width, double height, boolean wParent)

    {

	return Util.drawStringWrapped(g, extendedLabel(wParent),

					x+MARGIN, y+MARGIN, 

					width-MARGIN*2, height-MARGIN*2, 

					true, false);

    }



    public void drawLabel(Graphics g, int type, boolean wParent) {

	if (getParent() == null)

	    return; 



	// Draw our own label



	if (isLabelDrawable()) {

	    if (isOpen()) {

	        // top left corner 



	        drawTopLeftLabel(g);

	    }

	    else {

	        // centered, perhaps multi-line



		g.setColor(getLabelColour());



		setFont(g, type);



	        fullLabel = doDrawLabel(g, x(), y(), width(), height(), wParent);

            }

	} else {

	    fullLabel = false;

	}

    }



    public void drawLabel(Graphics g) {

	drawLabel(g, REG_FONT, false);

    }



    public void drawLabels(Graphics g) {



	if (width() < MIN_WIDTH || height() < MIN_HEIGHT) {

	    fullLabel = false;

	    return;

	}



	// Call method on each child



	if (isOpen()) {

	    Enumeration en = getChildren();



	    while (en.hasMoreElements()) {

	        EntityInstance e = (EntityInstance) en.nextElement();



                e.drawLabels(g);

            }

	}



	drawLabel(g);

    }



    public Dimension getLabelDim(Graphics g, int type, boolean wParent) {

	setFont(g, type);



	FontMetrics fm = g.getFontMetrics();



	int h = Util.fontHeight(fm);

	int w;



	if (wParent) {

	    EntityInstance pe = getEnterableParent();



	    if (pe != null) {

		w = Math.max(fm.stringWidth(pe.getLabel() + " | "), 

			     fm.stringWidth(label));

		h += h/2;

	    }

	    else {

		w = fm.stringWidth(label);

	    }

	}

	else {

	    w = fm.stringWidth(label);

	}



	return new Dimension(w, h);

    }



    public Dimension getLabelDim(Graphics g, int type) {

	return getLabelDim(g, type, false);

    }



    public Dimension getFitDim(Graphics g, int type, boolean wParent) {

	Dimension dim = getLabelDim(g, type, wParent);



	if (hasChildren() && !isOpen()) {

	    dim.width += CONTENTS_FLAG_DIM;

	}



	dim.width += MARGIN*3;

	dim.height += (MARGIN*3)/2;

	

	return dim;

    }



    public Dimension getFitDim(Graphics g, int type) {

	return getFitDim(g, type, false);

    }





    // Flags



    public void drawClientFlag(Graphics g) {

	// Draw flag that shows client exists



	if (drewClientFlag)

	    return;



	RealPoint p1 = 

	    new RealPoint(x() + width() - 5, y() + Util.ARROW_L);



	RealPoint p2 =

	    new RealPoint(x() + width() - Util.ARROW_L*2 - 5, y() + Util.ARROW_L);



	Util.drawArrow(g, p1, p2);



	drewClientFlag = true;

    }



    public void drawSupplierFlag(Graphics g) {

	// Draw flag that shows supplier exists



	if (drewSupplierFlag)

	    return;



	RealPoint p1 = 

	    new RealPoint(x() + width() - Util.ARROW_L*2 - 5, y() + Util.ARROW_L);



	RealPoint p2 =

		new RealPoint(x() + width() - 5, y() + Util.ARROW_L);



	Util.drawArrow(g, p1, p2);



	drewSupplierFlag = true;

    }



    public void drawContentsFlag(Graphics g) {

	if (width() < CONTENTS_FLAG_DIM*2 || height() < CONTENTS_FLAG_DIM*2)

	    return; 



	g.setColor(colour.darker());



	ScreenPoint pt0 = new ScreenPoint(x()+3, y()+3);



	g.drawRect(pt0.x, pt0.y, CONTENTS_FLAG_DIM, CONTENTS_FLAG_DIM);



	ScreenPoint pt1 = new ScreenPoint(pt0.x+2, pt0.y+4);

	ScreenPoint pt2 = new ScreenPoint(pt0.x+6, pt0.y+4);



	// g.setColor(Color.black);

	g.drawLine(pt1.x, pt1.y, pt2.x, pt2.y);



	pt1.set(pt0.x+4, pt0.y+2);

	pt2.set(pt0.x+4, pt0.y+6);



	g.drawLine(pt1.x, pt1.y, pt2.x, pt2.y);

    }



    // Edges



    public void drawEdges(Graphics g) { 

	EntityInstance ve, pe;

	ve = getVisibleEntity();

	if (ve == null) {
	    pe = null;
	} else {
	    pe = ve.getParent();
	}

	// Drill down to draw edges from each child

	Enumeration en = getChildren();

	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();

            e.drawEdges(g);

        }

	// All that follows seems to depend on pe not being null

	if (pe == null) {
	    return;
	}

	// Now draw edges from this entity



	EntityInstance root = dg.getRoot();



	en = srcRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    if (ri.getSrc() != this) {

		System.out.println("Error!!!!!");

	    }



	    if (ri.draw(g, true)) { // allow elision

	        EntityInstance dst = ri.getDst();


	        if (!root.descendent(dst) && 

		    pe.isSupplierRelationElided(ri.getRelationClass()))

	        {

		    ve.drawSupplierFlag(g);

	        }

	    }

	}



	// There may be edges that end with us that have no source

	// in the current diagram root. 



	en = dstRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    EntityInstance src = ri.getSrc();



	    if (!root.descendent(src)) {

	        if (ri.draw(g, true)) {	// allow elision

	            if (pe.isClientRelationElided(ri.getRelationClass())) {

	                ve.drawClientFlag(g);

		    }

		}

	    }

	}

    }



    public void drawHighlightedEdges(Graphics g) { 

	// Draw our edges (edges starting with us)



	EntityInstance root = dg.getRoot();



	Enumeration en = srcRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    if (ri.getHighlightFlag() && ri.getRelationClass().isActive()) {

	        ri.draw(g, false);	// no elision

	    }

	}



	en = dstRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    EntityInstance src = ri.getSrc();



	    if (!root.descendent(src) && ri.getHighlightFlag() &&

			ri.getRelationClass().isActive())

	    {

	        ri.draw(g, false);	// no elision

	    }

	}





	// Call method on each child



	en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



            e.drawHighlightedEdges(g);

        }

    }



    protected void doDrawCardinals(Graphics g, FontMetrics fm, int ht) {

	for (int i=0; i<dstCardinals.length; i++) {

	    RelationClass rc = dg.numToRelationClass(i);



	    if (dstCardinals[i] > 0 && 

		(!dg.allowElision() || rc.isVisible()))

	    {

		String cs = "" + dstCardinals[i];



		int w = fm.stringWidth(cs);



		double f = rc.getIOfactor();

	

		ScreenPoint pt = 

			new ScreenPoint(x()+width()*f-w/2, y()+height()+MARGIN+ht);



		g.setColor(rc.getColor());



		if (g instanceof PSGr) {

		    g.fillOval(pt.x-MARGIN, pt.y-ht+3, 

				Math.max(ht, w+MARGIN*2), ht);

		}

		else {

		    g.fillOval(pt.x-MARGIN, pt.y-ht+3, 

				Math.max(ht, w+MARGIN*2)+1, ht+1);

		}

	

		g.setColor(Color.black);



		g.drawOval(pt.x-MARGIN, pt.y-ht+3, 

				Math.max(ht, w+MARGIN*2), ht);



		g.setColor(CARD_FORE_COLOUR);



	        g.drawString(cs, pt.x, pt.y);

	    }

	}



	if (!isOpen())

	    return;



	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.doDrawCardinals(g, fm, ht);

	}

    }



    public void drawCardinals(Graphics g) {

	g.setFont(smallFont);



	FontMetrics fm = g.getFontMetrics();



	doDrawCardinals(g, fm, Util.fontHeight(fm));

    }





    public void getHighlightedEdges(Vector v) { 



	Enumeration en = srcRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    if (ri.getHighlightFlag() && ri.getRelationClass().isActive()) {

	        v.addElement(ri);

	    }

	}



	en = dstRelList.elements();



	EntityInstance root = dg.getRoot();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    EntityInstance src = ri.getSrc();



	    if (!root.descendent(src) && ri.getHighlightFlag() &&

			ri.getRelationClass().isActive())

	    {

		v.addElement(ri);

	    }

	}





	// Call method on each child



	en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.getHighlightedEdges(v);

        }

    }







    // draw a line segement between entity and coordinates



    public void drawSegment(Graphics gc, int x1, int y1) {

	int x0 = (int) Math.round(x() + width()/2);

	int y0 = (int) Math.round(y() + height()/2);



	gc.setColor(Color.black);

	gc.drawLine(x0, y0, x1, y1); 

    }



    // TA raw info routines



    public void getInstancesRaw(Vector v, Hashtable st) {

	// Get instance info for ourself, and then all our children



	if (parentClass != null) {

	    // Not $ROOT 



	    Integer[] pair = new Integer[2];



	    pair[0] = (Integer) st.get(id);

	    pair[1] = (Integer) st.get(parentClass.getId());



	    v.addElement(pair);

	}



	Enumeration en = getChildren();



	while(en.hasMoreElements()) {

	    EntityInstance child = (EntityInstance) en.nextElement();



	    child.getInstancesRaw(v, st);

	}

    }



    public void getRelationsRaw(Vector v, Hashtable st, Integer cid) {

	// Get the relations on us, and then all our children



	Enumeration en = srcRelList.elements(); 



	while(en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    Integer[] triple = new Integer[3];



	    RelationClass rc  = ri.getRelationClass();

	    EntityInstance e1 = ri.getSrc();

	    EntityInstance e2 = ri.getDst();



	    triple[0] = (Integer) st.get(rc.getId());

	    triple[1] = (Integer) st.get(e1.getId());

	    triple[2] = (Integer) st.get(e2.getId());



	    v.addElement(triple);

	} 		



	en = getChildren();



	while(en.hasMoreElements()) {

	    EntityInstance child = (EntityInstance) en.nextElement();



	    if (parentClass != null) {

		// If not root, output contain relation



	        Integer[] triple = new Integer[3];



	        triple[0] = cid;

		triple[1] = (Integer) st.get(id);

		triple[2] = (Integer) st.get(child.id);



	        v.addElement(triple);

	    }



	    child.getRelationsRaw(v, st, cid);

	}

    }





    // TA file output routines 



    public void writeInstances(PrintStream ps) throws IOException {

	// Write an instance line for ourself, and then all our children



	if (parentClass != null) {

	    // Not $ROOT 



	    ps.print(Diagram.INSTANCE_ID + " " + 

			qt(id) + " " + qt(parentClass.getId()) + "\n");

	}



	Enumeration en = getChildren();



	while(en.hasMoreElements()) {

	    EntityInstance child = (EntityInstance) en.nextElement();



	    child.writeInstances(ps);

	}

    }



    public void writeRelations(PrintStream ps) throws IOException {

	// Write out the relations on us, and then all our children



	Enumeration en = srcRelList.elements(); 



	while(en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    ri.writeRelation(ps); 

	} 		



	en = getChildren();



	while(en.hasMoreElements()) {

	    EntityInstance child = (EntityInstance) en.nextElement();



	    if (parentClass != null) {

		// If not root, output contain relation



		ps.print(Diagram.CONTAIN_ID + " " +

				qt(id) + " " + qt(child.id) + "\n"); 

	    }



	    child.writeRelations(ps);

	}

    }



    protected void writeLayout(PrintStream ps) throws IOException {

	// Write out layout attributes



	ps.print(Attribute.indent + "x = " + xLocal + "\n");

	ps.print(Attribute.indent + "y = " + yLocal + "\n"); 

	ps.print(Attribute.indent + "width = " + widthLocal + "\n"); 

	ps.print(Attribute.indent + "height = " + heightLocal + "\n");

    }



    public void

    writeLayoutAttributes(PrintStream ps) throws IOException {

        // Write out attribute record for us, and then our children



        ps.print(qt(id) + " {\n");



	writeLayout(ps); 



        ps.print("}\n\n");



        // Recurse for contained children



        Enumeration en = getChildren();



        while(en.hasMoreElements()) {

            EntityInstance child = (EntityInstance) en.nextElement();



            child.writeLayoutAttributes(ps);

        }

    }



    protected void writeElision(PrintStream ps, Vector ev, String id) {

	if (ev.size() > 0) {

	    ps.print(Attribute.indent + id + " = (");

	       

	    Enumeration en = ev.elements();



	    while(en.hasMoreElements()) {

		String rcId = (String) en.nextElement(); 



		if (en.hasMoreElements())

		    ps.print(rcId + " ");

		else

		    ps.print(rcId); 

	    }

	    ps.print(")\n"); 

	}

    }



    protected void regRawElision(Vector v, Hashtable st, Vector ev, String id) 

    {

	if (ev.size() > 0) {

	    int len = ev.size();

	    String[] sl = new String[len];



	    for (int i=0; i<len; i++) {

		sl[i] = (String) ev.elementAt(i);

	    }



	    regRawAttribute(v, st, id, sl);

	}

    }







    public void

    writeAttributes(PrintStream ps, boolean outputLyt) throws IOException

    {  

	// Write out attribute record for us, and then our children



	ps.print(qt(id) + " {\n");



	if (outputLyt)  

	    writeLayout(ps);



	if (label != null && !label.equals(mungeId(id))) {

	    // Save only if it differs from munged id (ie. auto-label)



	    ps.print(Attribute.indent + LABEL_ID + " = " +

			qt(label) + "\n");

	}



	if (description != null) {

	    ps.print(Attribute.indent + DESC_ID + " = \"" + 

			description + "\"\n");

	}



	if (parentClass != null) {

	    // First order attributes not associated with $ROOT 



	    Color pc = parentClass.getColor();



	    if (!colour.equals(pc)) {

	        int red = colour.getRed();

	        int green = colour.getGreen();

	        int blue = colour.getBlue();



	        ps.print(Attribute.indent + COLOUR_ID + " = (" +

		    ((double) red)/((double) MAX_RGB) + " " +

		    ((double) green)/((double) MAX_RGB) + " " +

		    ((double) blue)/((double) MAX_RGB) + ")\n"); 

	    }



	    writeElision(ps, inElision, IN_ELISION_ID);

	    writeElision(ps, outElision, OUT_ELISION_ID);

	    writeElision(ps, clientElision, CLIENT_ELISION_ID);

	    writeElision(ps, supplierElision, SUPPLIER_ELISION_ID);

	    writeElision(ps, internalElision, INTERNAL_ELISION_ID);



	    writeIOpoints(ps);



	    if (fontDelta != 0) {

		ps.print(Attribute.indent + FONTDELTA_ID + " = ");

		ps.print(fontDelta + "\n");

	    }

	}



	// Finally output the second-class attributes 



	if (attributes.size() > 0) 

	    super.writeAttributes(ps, false); 



	// End the record 

	ps.print("}\n\n"); 



	// Write any attributes of src relations



	Enumeration en = srcRelList.elements();



	while(en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    ri.writeAttributes(ps);

	}



	// Recurse for contained children



        en = getChildren();



        while(en.hasMoreElements()) {

            EntityInstance child = (EntityInstance) en.nextElement();



            child.writeAttributes(ps, outputLyt);

        }

    }



    public void getAttributesRaw(Vector mv, Hashtable st) {

	Vector v = new Vector();

 

	regRawAttribute(v, st, XPOSITION_ID, x());

	regRawAttribute(v, st, YPOSITION_ID, y());

	regRawAttribute(v, st, WIDTH_ID, width());

	regRawAttribute(v, st, HEIGHT_ID, height());



	if (label != null && !label.equals(mungeId(id))) {

	    // Save only if it differs from munged id (ie. auto-label)



	    regRawAttribute(v, st, LABEL_ID, label);

	}



	if (description != null) {

	    regRawAttribute(v, st, DESC_ID, description);

	}



	if (parentClass != null) {

	    // First order attributes not associated with $ROOT 



	    Color pc = parentClass.getColor();



	    if (!colour.equals(pc)) {

		regRawAttribute(v, st, COLOUR_ID, colour);

	    }



	    pc = parentClass.getLabelColor();



	    if (!labelColour.equals(pc)) {

		regRawAttribute(v, st, LABEL_COLOUR_ID, labelColour);

	    }



	    regRawElision(v, st, inElision, IN_ELISION_ID);

	    regRawElision(v, st, clientElision, OUT_ELISION_ID);

	    regRawElision(v, st, supplierElision, SUPPLIER_ELISION_ID);

	    regRawElision(v, st, internalElision, INTERNAL_ELISION_ID);



	    // addRawIOpoints(v, st);



	    if (fontDelta != 0) {

		regRawAttribute(v, st, FONTDELTA_ID, fontDelta);

	    }

	}



	// Finally output the second-class attributes 



	if (attributes.size() > 0) 

	    super.getAttributesRaw(v, st, false);



	Object[] obj = new Object[2];



	obj[0] = st.get(id);

	obj[1] = v;



	mv.addElement(obj);



	Enumeration en;



/*

	// Write any attributes of src relations



	Enumeration en = srcRelList.elements();



	while(en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    ri.writeAttributes(ps);

	}

*/

	// Recurse for contained children



        en = getChildren();



        while(en.hasMoreElements()) {

            EntityInstance child = (EntityInstance) en.nextElement();



            child.getAttributesRaw(mv, st);

        }

    }





    // Accessor functions and queries 



    public boolean canOpen() {

	return !containsList.isEmpty();

    }



    public boolean isOpen() {

	if (tmpOpen != NOT_USED)

	    return (tmpOpen == OPEN);



	return ((dg.getRoot() == this) ||

		(!isInRelationElided(Diagram.CONTAIN_ID) && canOpen()));

    }



    public boolean isOpenElided() {

	 return	!isInRelationElided(Diagram.CONTAIN_ID);

    }



    public EntityInstance getParent() { 

	return containedBy;

    }	



    public EntityInstance getEnterableParent() {

	if (containedBy == null || containedBy.isEnterable())

	    return containedBy;

	

	return containedBy.getEnterableParent();

    }



    public EntityClass getEntityClass() {

	return parentClass;

    }



    public boolean isRoot() {

	return (containedBy == null);

    }



    public boolean hasChildren() {

	return !containsList.isEmpty();

    }



    public int numChildren() {

	return containsList.size();

    }



    public Enumeration getChildren() {

	return containsList.elements();

    }



    public void resetCardinals(int numRelations) {

	dstCardinals = new int[numRelations];



	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.resetCardinals(numRelations);

	}

    }



    public void incCardinal(int ind) {

	dstCardinals[ind]++;

    }



    public void calcEdgeCardinals(boolean hOnly) {

	Enumeration en = srcRelList.elements();



	EntityInstance ve = getVisibleEntity();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();

	   

	    int ind = ri.getRelationClass().getNid();



	    EntityInstance vsrc = ri.getDst().getVisibleEntity();



	    if (vsrc != null && vsrc != ve && (!hOnly || ri.getHighlightFlag()))

	        vsrc.incCardinal(ind);

	}



	EntityInstance root = dg.getRoot();



	en = dstRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    if (!root.descendent(ri.getSrc())) {

	        int ind = ri.getRelationClass().getNid();



		if (!hOnly || ri.getHighlightFlag())

	            ve.incCardinal(ind);

	    }

	}



	en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.calcEdgeCardinals(hOnly);

	}

    }

    



    public void getClients(Vector v, EntityInstance root) {

	Enumeration en = dstRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    EntityInstance src = ri.getSrc();



	    if ((this == getVisibleEntity(root) || 

		!getParent().isInRelationElided(AGG_ELISION_ID)) &&

	        !root.descendent(src) && !v.contains(src))

	    {

		v.addElement(src);

	    }

	}



	en = containsList.elements();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.getClients(v, root);

	}

    }



    public void getSuppliers(Vector v, EntityInstance root) {

	Enumeration en = srcRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    EntityInstance dst  = ri.getDst();

	    EntityInstance vsrc = getVisibleEntity(root);



	    if ((this == vsrc || !getParent().isInRelationElided(AGG_ELISION_ID)) && !root.descendent(dst) && !v.contains(dst))

	    {

		v.addElement(dst);

	    }

	}



	en = containsList.elements();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.getSuppliers(v, root);

	}

    }



    public String toString() {

	return label;

    }



    public Diagram getDiagram() {

	return dg;

    }



    public String getDescription() {

	if (description != null)

	    return description; 

	else

	    return "";  

    }



    public void setDescription(String str) {

	description = str;

    }



    public void setLabel(String str) {

	label = str;

    }



    public Layout getLayout() {

/*

	if (dg.getTopInstance() == this) {

	    return dg.calcRootLayout();

	}

*/

	return new Layout(x(), y(), width(), height());

    }



    public Layout getLocalLayout() {

	return new Layout(xLocal, yLocal, widthLocal, heightLocal);

    }



    public void setGlobalLayout(Layout lyt) {

	setX(lyt.x);

        setY(lyt.y);

	setWidth(lyt.width);

	setHeight(lyt.height);

    } 



    public void setLayout(Layout lyt) {

	// Layout is global, adjust it to coords local to parent


	double xScale = dg.getScaleX();

	double yScale = dg.getScaleY();


	if (parentClass != null && containedBy != null) {

	    Layout pLyt = containedBy.getLayout(); 



	    xLocal = (lyt.x - pLyt.x)/xScale; 

	    yLocal = (lyt.y - pLyt.y)/yScale;

	}



	widthLocal = lyt.width/xScale;

	heightLocal = lyt.height/yScale; 



	setGlobalLayout(lyt);

    } 



    public void setAbsoluteLayout(Layout lyt) {

	// Layout is global, adjust it to coords local to parent



	xLocal = lyt.x;

	yLocal = lyt.y;

	

	// $ROOT always remains at 0, 0 



	widthLocal = lyt.width;

	heightLocal = lyt.height;



	setGlobalLayout(lyt);

    } 



    public void saveLayout() {

	tempLyt = new Layout(x(), y(), width(), height());



	Enumeration en = getChildren();



        while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



            e.saveLayout();

        }

    }



    public void restoreLayout() {

	if (tempLyt == null) {

	    System.out.println("Layout: restore without save " + label);

	    return;

	}



	setGlobalLayout(tempLyt);

	tempLyt = null;



	Enumeration en = getChildren();



        while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



            e.restoreLayout();

        }

    }





    public void moveBend(Bend bend, double xScale, double yScale) {

	bend.xLocal = (bend.x - x())/xScale;

	bend.yLocal = (bend.y - y())/yScale;

    }



    public void addBend(Bend bend, double xScale, double yScale) {

	moveBend(bend, xScale, yScale);

	if (!bends.contains(bend)) {

	    bends.addElement(bend);

	}

    }



    public void addBend(Bend bend) {

	if (!bends.contains(bend)) {

	    bends.addElement(bend);

	}

    }



    public void deleteBend(Bend bend) {

	bends.removeElement(bend);

    }



    public Layout getChildBoundingBoxLocal() {

	// Calc the bounding box of children



	if (containsList.isEmpty())

	    return null;



	Enumeration en = getChildren();



        double x1 = Double.MAX_VALUE;	// Becomes the smallest child.xLocal

        double x2 = Double.MIN_VALUE;	// Becomes largest child.(xLocal + widthLocal);

        double y1 = Double.MAX_VALUE;	// Becomes the smallest child.yLocal

        double y2 = Double.MIN_VALUE;	// Becomes the largest child.(yLocal+heightLocal);



        while (en.hasMoreElements()) {

            EntityInstance e = (EntityInstance) en.nextElement();



            if (e.xLocal < x1)

                x1 = e.xLocal;



            if (e.yLocal < y1)

                y1 = e.yLocal;



            if (e.xLocal + e.widthLocal > x2)

                x2 = e.xLocal + e.widthLocal;



            if (e.yLocal + e.heightLocal > y2)

                y2 = e.yLocal + e.heightLocal;

        }



	return new Layout(x1, y1, x2-x1, y2-y1);

    }



    public Layout getChildBoundingBox() {

        // Calc the global bounding box of children



        if (containsList.isEmpty())

            return null;



        Enumeration en = getChildren();



        double x1 = Double.MAX_VALUE;

        double x2 = Double.MIN_VALUE;

        double y1 = Double.MAX_VALUE;

        double y2 = Double.MIN_VALUE;



        while (en.hasMoreElements()) {

            EntityInstance e = (EntityInstance) en.nextElement();



            if (e.x() < x1)

                x1 = e.x();



            if (e.y() < y1)

                y1 = e.y();



            if (e.x() + e.width() > x2)

                x2 = e.x() + e.width();



            if (e.y() + e.height() > y2)

                y2 = e.y() + e.height();

        }



        return new Layout(x1, y1, x2-x1, y2-y1);

    }



    public void translateChildrenLocal(double dx, double dy) {

        Enumeration en = getChildren();



	while (en.hasMoreElements()) {

            EntityInstance e = (EntityInstance) en.nextElement();

	    e.xLocal += dx;

	    e.yLocal += dy;

	}



	en = bends.elements();



	while (en.hasMoreElements()) {

	    Bend bend = (Bend) en.nextElement();

	    bend.xLocal += dx;

	    bend.yLocal += dy;

	}

    }


    public Color getBackgroundWhenOpen() {


        // Container is handled specially
        // Find colour from depth



        float v = Diagram.BG - 0.05f;

        EntityInstance b = this;



        if (isTransparent()) {
	    b = b.containedBy;
	}


        EntityInstance root = dg.getRoot();



        if (!root.descendent(this)) {
  	    root = root.common(this);
	}


	while(b != root) {

	    v -= 0.05f;

	    b = b.containedBy; 
	}



	Color c = new Color(v, v, v);



	if (Util.isBlack(c)) {

	    c = Color.lightGray;
	}

	return c;
    }
	    

    public Color getBackground() {

	Color c;

	if (isOpen()) {
	    c = getBackgroundWhenOpen();
	} else {
	    c = colour;
	}
	return c;
    }



    public boolean isRelationElided(Vector v, String id) {

	String str = (String) est.get(id);



	//if (str == null)

	    //return false;

        // [irbull] commented this out because if an attributre was not present when it was first created, it would never use it.



	return v.contains(id);

    }



    protected boolean toggleElision(Vector v, String id) {

	if (isRelationElided(v, id)) {

	    if (!v.removeElement(id)) {

		MsgOut.println("remove() failed");

	    }

	    return false; 

	} 

	else {

	    v.addElement(id); 

	    return true; 

	}

    }

 

    public boolean toggleInElision(String id) {

	MsgOut.dprintln("toggle in " + id);

	return toggleElision(inElision, id);

    }

 

    public boolean toggleInElision(RelationClass rc) {

        return toggleInElision(rc.getId());

    }



    public boolean toggleOutElision(String id) {

	MsgOut.dprintln("toggle out " + id);



	return toggleElision(outElision, id);

    }



    public boolean toggleOutElision(RelationClass rc) {

	return toggleOutElision(rc.getId());

    }



    public boolean toggleClientElision(String id) {

	MsgOut.dprintln("toggle client " + id);



	return toggleElision(clientElision, id);

    }



    public boolean toggleClientElision(RelationClass rc) {

	return toggleClientElision(rc.getId());

    }



    public boolean toggleSupplierElision(String id) {

	MsgOut.dprintln("toggle supplier " + id);



	return toggleElision(supplierElision, id);

    }



    public boolean toggleSupplierElision(RelationClass rc) {

	return toggleSupplierElision(rc.getId());

    }



    public boolean toggleInternalElision(String id) {

	MsgOut.dprintln("toggle internal " + id);



	return toggleElision(internalElision, id);

    }



    public boolean toggleInternalElision(RelationClass rc) {

	return toggleInternalElision(rc.getId());

    }



    public void setTempOpenFlag(int state) {

	tmpOpen = state;

    }



    // Queries 



    public boolean hasId(String id) {

	return this.id.equals(id);

    }



    public boolean isInRelationElided(RelationClass rc) {

        boolean tmp = isRelationElided( inElision, rc.getId() );

	return tmp;

    }



    public boolean isInRelationElided(String id) {

	return isRelationElided(inElision, id);

    }



    public boolean isOutRelationElided(RelationClass rc) {

	return isRelationElided(outElision, rc.getId());

    }



    public boolean isOutRelationElided(String id) {

	return isRelationElided(outElision, id);

    }



    public boolean isClientRelationElided(RelationClass rc) {

	return isRelationElided(clientElision, rc.getId());

    }



    public boolean isClientRelationElided(String id) {

	return isRelationElided(clientElision, id);

    }



    public boolean isSupplierRelationElided(RelationClass rc) {

	return isRelationElided(supplierElision, rc.getId());

    }



    public boolean isSupplierRelationElided(String id) {

	return isRelationElided(supplierElision, id);

    }



    public boolean isInternalRelationElided(RelationClass rc) {

	return isRelationElided(internalElision, rc.getId());

    }



    public boolean isInternalRelationElided(String id) {

	return isRelationElided(internalElision, id);

    }



    public boolean isContainedBy(EntityInstance e2) {

	EntityInstance c = containedBy;



	while (c != null) {

	    if (c == e2)

		return true;



	    c = c.containedBy;

	}



	return false;

    }



    public boolean descendent(EntityInstance e) {

	EntityInstance ce = e;



	while (ce != null) {

	    if (ce == this) {

		return true;

	    }

	    ce = ce.containedBy;

	}



	return false;

    }



    // Return the common parent entity



    public EntityInstance common(EntityInstance e) {

	EntityInstance e1 = this.containedBy;



	if (e1 == null)

	    return e1; // All things have root as containment ancestor



	while(e1 != null) {

	    EntityInstance e2 = e;



	    while(e2 != null) {

		if (e1 == e2)

		    return e1;

		e2 = e2.containedBy;

	    }



	    e1 = e1.containedBy;

	}



	MsgOut.println("EntityInstance.common: failed");



	return null;

    }

    public EntityInstance getVisibleEntity(EntityInstance root) {

	if (this == root) {
	    return root;
	}

	if (visibleEntityCache == null) {

	    // Calculate it once per draw

	    EntityInstance e = this;

//	    if (!dg.isClientOrSupplier(e) || true) {

		EntityInstance ce = e;
		EntityInstance top;

		if (root.descendent(e)) {
		    top = root;
		} else {
		    top = dg.getClientOrSupplier(e);
		    if (top == null) {
			return null;
		    }
		    top = top.getParent();
		}

	        do {
		    ce = ce.containedBy;

		    if (ce == null) {
			return e;
		    }

		    if (ce != top && !ce.isOpen()) {
			e = ce;
		    }
		} while (ce != top);
//	    }
	    visibleEntityCache = e;
	}
	return visibleEntityCache;

    }



    public EntityInstance getVisibleEntity() {

	return getVisibleEntity(dg.getRoot());

    }





    // Return true if we're a client of passed entity.

    // That is, a relation exist between us, or our children and 

    // the passed entity or its children.



    public boolean isClientOf(EntityInstance pc) {

	Enumeration en = srcRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    EntityInstance e = ri.getDst();



	    if (e == pc)

		return true;



	    if (pc.isContainedBy(e))

		return true;

	}



	en = getChildren();



	while(en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    if (e.isClientOf(pc))

		return true;

	}



	return false;

    }





    // Return reference to entity the cursor is over. 

    // Return null if none.



    public EntityInstance getMouseOver(int x, int y, boolean annotActive) {

	if (annotActive) {

	    // Check to see if it is over a tab



	    if (getMouseOverTab(x, y) != NO_TAB) {

		return this; 

	    }

	}



	if (isPointOver(x, y)) {

	    if (isOpen()) {

	        Enumeration en = getChildren();



	        while(en.hasMoreElements()) {

		    EntityInstance e = (EntityInstance) en.nextElement();



		    EntityInstance over = e.getMouseOver(x, y, annotActive);



	            if (over != null)

	                return over; 

	        }

	    }



	    // It's not over any of our children, so it's over us.

	    

	    return this; 

	}



	return null;

    }



    public Bend getMouseOverBend(int x, int y) {

	Enumeration en = bends.elements();



	while (en.hasMoreElements()) {

	    Bend bend = (Bend) en.nextElement();



	    if (bend.isOver(x, y)) 

		return bend;

	}



	en = getChildren();



	while(en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    if (e.isOpen()) {

	        Bend bend = e.getMouseOverBend(x, y);



	        if (bend != null)

	            return bend; 

	    }

	}

	return null;

    }



    public int getMouseOverTab(int x, int y) {

	for (int i=PUBLIC_TAB; i<=PRIVATE_TAB; i++) {

	    if (tabStatus[i] != ANNOT_NONE && isPointOverTab(x, y, i)) {

		return i;

	    }

	}

	return NO_TAB;

    }



    public RelationInstance getMouseOverEdge(int x, int y, int timeStamp) {

	// if (isOpen()) {

	    Enumeration en1 = getChildren();



	    while(en1.hasMoreElements()) {

		// Look at each child
		EntityInstance ce = (EntityInstance) en1.nextElement();



		Enumeration en2 = ce.dstRelationElements();



		RelationInstance ri;



		while (en2.hasMoreElements()) {

		    ri = (RelationInstance) en2.nextElement();



		    if (ri.isNearEdge(x, y, MOUSE_NEAR_EDGE_THRESHOLD, timeStamp))

		    {

			return ri;

		    }

		}



		en2 = ce.srcRelationElements();



		while (en2.hasMoreElements()) {

		    ri = (RelationInstance) en2.nextElement();



		    if (ri.isNearEdge(x, y, MOUSE_NEAR_EDGE_THRESHOLD, timeStamp))

		    {

			return ri;

		    }

		}



		ri = ce.getMouseOverEdge(x, y, timeStamp);



		if (ri != null)

		    return ri;

	    }

	// }

	return null;

    }



    public void getMouseOverAllEdges(int x, int y, Vector v) {

	Enumeration en = dstRelationElements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    if (ri.isNearEdge(x, y, MOUSE_NEAR_EDGE_THRESHOLD, -1)) {

		if (!v.contains(ri))

		    v.addElement(ri);

	    }

	}



	en = srcRelationElements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    if (ri.isNearEdge(x, y, MOUSE_NEAR_EDGE_THRESHOLD, -1)) {

		if (!v.contains(ri))

		    v.addElement(ri);

	    }

	}

    }





    public EntityInstance intersects(Layout lyt) {

	if (containsLayout(lyt)) {

	    if (!containsList.isEmpty()) {

		// Check children



		Enumeration en = getChildren();



		while(en.hasMoreElements()) {

		    EntityInstance e = (EntityInstance) en.nextElement();



		    EntityInstance oe = e.intersects(lyt); 



		    if (oe != null)

			return oe; 

		}



		return null; 

	    }

	    return this;  // over us 

	}

	else if (intersectsLayout(lyt)) 

	    return this;



	return null; 

    }





    public EntityInstance containing(Layout lyt) {

        if (containsLayout(lyt)) {

            if (!containsList.isEmpty() && isOpen()) {

                // Check children



                Enumeration en = getChildren();



                while(en.hasMoreElements()) {

                    EntityInstance e = (EntityInstance) en.nextElement();



                    EntityInstance oe = e.containing(lyt);



                    if (oe != null)

                        return oe;

                }

            }

            return this;

        }

        return null;

    }



    public Vector getCoverList(Layout coverLyt) {

	// Return the set of children contained in layout



	Vector contents = new Vector();



	Enumeration en = getChildren();



	while(en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    Layout lyt = e.getLayout();



	    if (coverLyt.x < lyt.x && lyt.x+lyt.width < coverLyt.x+width() &&

		coverLyt.y < lyt.y && lyt.y+lyt.height < coverLyt.y+height())

	    {

		contents.addElement(e);

	    }

	}



	return contents;

    }



    public Enumeration srcRelationElements() {

	return srcRelList.elements(); 

    }



    public Enumeration dstRelationElements() {

	return dstRelList.elements(); 

    }



    protected void addSrcRels(Vector v) {

	Enumeration en = srcRelList.elements();



	while(en.hasMoreElements()) {

	    v.addElement(en.nextElement());

	}



	en = getChildren();



	while(en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.addSrcRels(v);

	}

    }



    public Enumeration srcRelationAbstractedElements() {

	// Return enumeration for all edges which

	// have this entity as the source (real and abstracted).



	Vector v = new Vector();



	addSrcRels(v);



	return v.elements();

    }



    protected void addDstRels(Vector v) {

	Enumeration en = dstRelList.elements();



	while(en.hasMoreElements()) {

	    v.addElement(en.nextElement());

	}



	en = getChildren();



	while(en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.addDstRels(v);

	}

    }



    public Enumeration dstRelationAbstractedElements() {

	// Return enumeration for all edges which

	// have this entity as the destination (real and abstracted).



	Vector v = new Vector();



	addDstRels(v);



	return v.elements();

    }



    public Vector groupRegion(Layout lyt) {

	// This entity isn't in the group, but any children which

	// intersect the passed region are.



	if (containsList.isEmpty())

	    return null;



	Vector grp = new Vector();



	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement(); 



	    if (e.intersectsLayout(lyt)) {

		e.setGroupFlag(); 

		grp.addElement(e);

	    }

	}



	if (grp.isEmpty())

	    return null;



	return grp;

    }



    public void getGroup(Vector grp) {

	if (getGroupFlag())

	    grp.addElement(this);



	if (containsList.isEmpty())

	    return;



	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement(); 



	    e.getGroup(grp);

	}

    }


    protected void setGroupFlag() {

       m_groupFlag = true; 
    }



    protected void clearGroupFlag() {

        m_groupFlag = false;
    }



    public boolean getGroupFlag() {

	return m_groupFlag;

    }



    protected void setGroupKeyFlag() {

	m_groupKeyFlag = true;

    }



    protected void clearGroupKeyFlag() {

	m_groupKeyFlag = false;

    }



    public boolean getGroupKeyFlag() {

	return m_groupKeyFlag;

    }



    public void getHighlightGroup(Vector grp) {

	if (highlightFlag)

	    grp.addElement(this);



	if (containsList.isEmpty())

	    return;



	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement(); 



	    e.getHighlightGroup(grp);

	}

    }



    public void setHighlightFlag() {

	highlightFlag = true;

    }



    public boolean getHighlightFlag() {

	return highlightFlag;

    }



    public void clearHighlightFlag() {

	highlightFlag = false;

    }



    public void clearAllFlags() {

	clearGroupFlag();

	clearGroupKeyFlag();

	highlightFlag = false;

	tmpOpen = NOT_USED;



	Enumeration en = srcRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    ri.clearHighlightFlag();

	    ri.clearGroupFlag();

	}



	en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement(); 



	    e.clearAllFlags();

	}

    }



    public void clearHighlightFlags() {

	highlightFlag = false;



	Enumeration en = srcRelList.elements();



	while (en.hasMoreElements()) {

	    RelationInstance ri = (RelationInstance) en.nextElement();



	    ri.clearHighlightFlag();

	}



	en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement(); 



	    e.clearHighlightFlags();

	}

    }



    public void clearGroupFlags() {

	clearGroupFlag();

	clearGroupKeyFlag();



	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement(); 



	    e.clearGroupFlags();

	}

    }



    public boolean isClippedLabel() {

	return !fullLabel;

    }



    // Internal attributes of different entities



    public boolean isTransparent() {

	return false;

    }



    public boolean isCloseable() {

	return true;

    }



    public boolean isClickable() {

	return true;

    }



    public boolean isLabelDrawable() {

	return true;

    }



    public boolean isEnterable() {

	return true;

    }



    public void incDstCount() {

	dstCount++;

    }



    public int getDstCount() {

	return dstCount;

    }



    public void setMark(int val) {

	mark = val;

    }



    public int getMark() {

	return mark;

    }



    public void clearMarks() {

	mark = 0;



	Enumeration en = getChildren();



	while (en.hasMoreElements()) {

	    EntityInstance e = (EntityInstance) en.nextElement();



	    e.clearMarks();

	}

    }



    static public Font getSmallFont() {

	return smallFont;

    }

    public double x() {
	return(m_x);
    }
    public double y() {
	return(m_y);
    }
    public double width() {
	return(m_width);
    }
    public double height() {
	return(m_height);
    }

    protected void setX(double x) {
	m_x = x;
    }
    protected void setY(double y) {
	m_y = y;
    }
    protected void setWidth(double width) {
	m_width = width;
    }
    protected void setHeight(double height) {
	m_height = height;
    }


}





