package lsedit;

import java.awt.*;

import java.util.*;

import java.io.*; 



public abstract class LandscapeObject extends Object {

    public final static String COLOUR_ID =           "color";

    public final static String LABEL_COLOUR_ID =     "labelcolor";

    

    protected static final int MAX_RGB = 255; 



    protected String id, label, description, title;



    protected int style = 0;	// Default style - semantics vary



    protected Diagram dg;	// Diagram which contains it



    protected int m_nid = -1;



    protected static int edge_mode = Diagram.BEST_EDGE;



    protected Color colour;

    protected Color labelColour;



    // Attribute database

    protected Hashtable attributes = new Hashtable(5);



    protected void saveForUndo(LandscapeObject undoObject) {

	undoObject.id = id;

	undoObject.label = label; 

	undoObject.dg = dg;

	undoObject.setNid(getNid());

	undoObject.attributes = (Hashtable) attributes.clone(); 

    }



    protected void undo(LandscapeObject undoObject) {

	id = undoObject.id;

	label = undoObject.label; 

	dg = undoObject.dg;

	setNid(undoObject.getNid());

	attributes = undoObject.attributes; 

    }



    protected String qt(String str) {

	return AttributeValueItem.qt(str);

    }



    abstract public boolean hasId(String id);



    // --------------

    // Public methods 

    // --------------



    public void regRawAttribute(Vector v, Hashtable st, String id, 

				Object type, Object val)

    {

	Object[] ent = new Object[3];



	ent[0] = st.get(id);

	ent[1] = type;

	ent[2] = val;



	v.addElement(ent);

    }



    public void regRawAttribute(Vector v, Hashtable st, String id, int val) {

	regRawAttribute(v, st, id, Attribute.INT, new Integer(val));

    }



    public void regRawAttribute(Vector v, Hashtable st, String id, double val) 

    {

	regRawAttribute(v, st, id, Attribute.DOUBLE, new Double(val));

    }



    public void regRawAttribute(Vector v, Hashtable st, String id, String val) 

    {

	regRawAttribute(v, st, id, Attribute.STRING, val);

    }



    public void regRawAttribute(Vector v, Hashtable st, 

					String id, int[] val)

    {

	regRawAttribute(v, st, id, Attribute.INT_LIST, val);

    }



    public void regRawAttribute(Vector v, Hashtable st, 

					String id, double[] val)

    {

	regRawAttribute(v, st, id, Attribute.DOUBLE_LIST, val);

    }



    public void regRawAttribute(Vector v, Hashtable st, 

					String id, String[] val)

    {

	regRawAttribute(v, st, id, Attribute.STRING_LIST, val);

    }



    public void regAttributeRaw(Vector v, Hashtable st, 

					Attribute attr, boolean classType)

    {

	if (attr.avi == null) {

	    if (classType) {

		regRawAttribute(v, st, attr.id, Attribute.NULL, null);

	    }

	}

	else if (classType || !attr.cloneOnAssign) {

	    if (attr.avi.next != null) {

		AttributeValueItem avi = attr.avi;

		

		int num = 0;



		while(avi != null) {

		    avi = avi.next;

		    num++;

		}



		String[] sl = new String[num];



		avi = attr.avi;



		for (int i=0; i<num; i++) {

		    sl[i] = avi.value;

		}



		regRawAttribute(v, st, attr.id, sl);

	    }

	    else {

		regRawAttribute(v, st, attr.id, attr.avi.value);

	    }

	}

    }



    public void regRawAttribute(Vector v, Hashtable st, String id, Color c) {

	int red = c.getRed();

        int green = c.getGreen();

        int blue = c.getBlue();



	double[] dl = new double[3];



	dl[0] = ((double) red)/((double) MAX_RGB);

	dl[1] = ((double) green)/((double) MAX_RGB);

	dl[2] = ((double) blue)/((double) MAX_RGB);



	regRawAttribute(v, st, id, dl);

    }





    // If a new attribute, add it to the attribute database,

    // otherwise merely update the value.



    public boolean processFirstOrder(Attribute attr) {

	if (attr.id.equals(COLOUR_ID)) {

            colour = attr.parseColour();

	    return true;

	}

	else if (attr.id.equals(LABEL_COLOUR_ID)) {

            labelColour = attr.parseColour();

	    return true;

	}

	return false;

    }



    public void addAttribute(Attribute attr) {

	if (processFirstOrder(attr)) 

	    return;



        Attribute curAttr = (Attribute) attributes.get(attr.id);



        if (curAttr == null) {

            // New attribute



            attributes.put(attr.id, attr);

        }

        else if (attr.avi != null) {

	    // Replace any old attributes with overrides



	    attributes.remove(curAttr);

	    attributes.put(attr.id, attr);

	}

    }



    public boolean addRawAttribute(String id) {

	Attribute attr = new Attribute(id, null);



	addAttribute(attr);



	return true;

    }



    public boolean addRawAttribute(String id, String val) {

	AttributeValueItem avi = new AttributeValueItem(val);



	Attribute attr = new Attribute(id, avi);



	addAttribute(attr);



	return true;

    }



    public boolean addRawAttribute(String id, double val) {

	System.out.println("0");

	return false;

    }



    public boolean addRawAttribute(String id, int val) {

	System.out.println("1");

	return false;

    }



    public boolean addRawAttribute(String id, int[] val) {

	System.out.println("2");

	return false;

    }



    public boolean addRawAttribute(String id, double[] val) {

	if (id.equals(COLOUR_ID)) {

	    colour = new Color((float) val[0], (float) val[1], (float) val[2]);



	    return true;

	}

	else if (id.equals(LABEL_COLOUR_ID)) {

	    labelColour = 

		new Color((float) val[0], (float) val[1], (float) val[2]);



	    return true;

	}

	return false;

    }



    public boolean addRawAttribute(String id, String[] val) {

	AttributeValueItem favi = null;

	AttributeValueItem lavi = null;



	int num = val.length;



	for (int i=0; i<num; i++) {

	    AttributeValueItem avi = new AttributeValueItem(val[i]);



	    if (lavi != null)

		lavi.next = avi;



	    if (favi == null)

		favi = avi;



	    lavi = avi;

	}



	Attribute attr = new Attribute(id, favi);

	addAttribute(attr);



	return true;

    }







    // Return the attribute with the associated id



    public Attribute getAttribute(String id) {

	return (Attribute) attributes.get(id);

    }



    public void replaceAttribute(Attribute newAttr) {

	addAttribute(newAttr);

    }



    public void writeAttributes(PrintStream ps, boolean classType)

				throws IOException 

    {

	// Output the body of the attribute record 

	// The child method outputs the header and tail 



	if (classType) {

	    if (colour != null) {

	        int red = colour.getRed();

	        int green = colour.getGreen();

	        int blue = colour.getBlue();



	        ps.print(Attribute.indent + COLOUR_ID + " = (" +

		    ((double) red)/((double) MAX_RGB) + " " +

		    ((double) green)/((double) MAX_RGB) + " " +

		    ((double) blue)/((double) MAX_RGB) + ")\n"); 

	    }



	    if (labelColour != null) { 

	        int red = labelColour.getRed();

	        int green = labelColour.getGreen();

	        int blue = labelColour.getBlue();



	        ps.print(Attribute.indent + LABEL_COLOUR_ID + " = (" +

		    ((double) red)/((double) MAX_RGB) + " " +

		    ((double) green)/((double) MAX_RGB) + " " +

		    ((double) blue)/((double) MAX_RGB) + ")\n"); 

	    }

	}



	Enumeration en = attributes.elements();



	while (en.hasMoreElements()) {

	    Attribute attr = (Attribute) en.nextElement();



	    attr.writeAttribute(ps, classType);

	}

    }



    public void getAttributesRaw(Vector v, Hashtable st, boolean classType) {

	if (classType) {

	    if (colour != null) {

	        regRawAttribute(v, st, COLOUR_ID, colour);

	    }

	    if (labelColour != null) {

	        regRawAttribute(v, st, LABEL_COLOUR_ID, labelColour);

	    }

	}



	Enumeration en = attributes.elements();



	while (en.hasMoreElements()) {

	    Attribute attr = (Attribute) en.nextElement();



	    regAttributeRaw(v, st, attr, classType);

	}

    }



    public boolean hasAttributes(boolean classType) {

	if (classType)

	    return true;



	Enumeration en = attributes.elements();



	while (en.hasMoreElements()) {

	    Attribute attr = (Attribute) en.nextElement();



	    if (!attr.cloneOnAssign)

		return true;

	}



	return false;

    }



    public String toString() {

	return id;

    }



    public String getId() {

	return id;

    }



    public void setId(String newId) {

	id = newId;

    }



    public String getLabel() {

	return label;

    }



    public void setLabel(String newLabel) {

	label = newLabel;

    }



    public String getDescription() {

	return description;

    }



    public void setDescription(String desc) {

	description = desc;

    }



    public String getTitle() {

	return title;

    }



    public void setTitle(String title) {

	this.title = title;

    }



    public int getNid() {

	return m_nid;

    }

  

    public void setNid(int nid) {

	this.m_nid = nid;

    }



    public int getStyle() {

	return style;

    }



    public Color getColor() {

	return colour;

    }



    public Color getLabelColor() {

	return labelColour;

    }



    public static void setEdgeMode(int mode) {

	edge_mode = mode;

    }



    public static int getEdgeMode() {

	return edge_mode;

    }

}

