package lsedit;

import java.applet.*;

import java.awt.*;

import java.util.*;



public class LegendBox extends TabbedBox {



    protected static final int CB_WD = 15;

    protected static final int CB_HT = 15;

    protected static final int WIDTH = 25;

    protected static final int HEIGHT = 20;



    protected static final Font textFont =

		new Font("Helvetica", Font.PLAIN, 11); 



    protected static final String endEntMsg =

	"Pass cursor over entity for description.";



    protected static final String endRelMsg1 =

	"Checkboxes select visible relations.";



    protected static final String endRelMsg2 =

	"Click relation arrow for description.";



    protected static final String helpStr =

	"This box shows the types of entities and relations that are present in the current landscape.";



    protected static final int MAX_NUM_CHECKS = 20;



    protected Checkbox[] relnChk = new Checkbox[MAX_NUM_CHECKS];

    protected RelationClass[] relnClass = new RelationClass[MAX_NUM_CHECKS];



    protected Vector objects;



    protected void drawContents(Diagram dg, Graphics g, boolean fullInfo) {

	Rectangle r = bounds();



        int xpos = GAP*2;

        int ypos = -scrollBar.getValue();

	int iypos = ypos;



       	g.setFont(textFont);



        FontMetrics fm = g.getFontMetrics();

        int h = Util.fontHeight(textFont); // Util.fontHeight(fm);



	int bh = h*2;

	int bw = (bh*4)/3;

	int tw = r.width-bw-GAP*3;



	g.setColor(Color.black);



	if (fullInfo) {

            ypos += h;

	    g.drawString("Entities", xpos, ypos);

	    g.drawLine(xpos, ypos+3, xpos+fm.stringWidth("Entities"), ypos+3);

	    ypos += h-2;

	}

	else {

	    ypos += 5;

	}



	// Draw legend for entities



	Enumeration en;



	int item = 0;

	int n = 0;



	if (fullInfo) {



	en = dg.enumEntityClasses();

	objects = new Vector();



	while(en.hasMoreElements()) {

	    EntityClass ec = (EntityClass) en.nextElement();



            String id = ec.getId();

	    String label = ec.getLabel();



	    if (!id.equals(EntityClass.ENTITY_BASE_CLASS_ID)) {

		if (true /* item >= itemNo && item < vitem */ ) {

	    	    EntityInstance e = ec.newEntity("A", dg);

		    Layout lyt = new Layout(xpos, ypos + 2, bw, bh);

		    e.setAbsoluteLayout(lyt);
		    e.draw(g);

		    Rectangle rect = new Rectangle(r.x + xpos, r.y+ypos+2, bw, h);
		    objects.addElement(new LegendObject(rect, ec.getDescription()));

		    g.setColor(Color.black);
        	    g.setFont(textFont);

	            Util.drawStringClipped(g, label, xpos+bw+GAP, ypos, tw, h); 
	            ypos += bh+GAP*2;
		}
		item++;
		n++;

	    }

	}



	if (n == 0) {

	    g.setColor(Color.blue);

	    g.draw3DRect(xpos, ypos, bw, h, true);

	    g.fill3DRect(xpos+1, ypos+1, bw-1, h-1, true);



	    g.setColor(Color.black);

	    Util.drawStringClipped(g, "Entity", xpos+bw+GAP, ypos, tw, h);

	    ypos += h+GAP;

	    item++;

	}



	ypos += (bh + h)/2;



	Util.drawStringClipped(g, endEntMsg, xpos, ypos, tw, h);



	ypos += 2*h;



	} // if (fullInfo)



        g.setColor(Color.black);

        g.drawString("Relations", xpos, ypos+h);

        ypos += h+3;

        g.drawLine(xpos, ypos, xpos+fm.stringWidth("Relations"), ypos);

        ypos += h/2;



	// Draw legend for relations



	en = dg.enumRelations();



	n = 0;

        while(en.hasMoreElements()) {

            RelationClass rc = (RelationClass) en.nextElement();



            String id = rc.getId();

	    String label = rc.getLabel();



	    if (!dg.excludeReln(rc)) {

		String str = label + " (" + (n+1) + ")";



		if (true /* item >= itemNo && item < vitem */ ) {



		    g.setColor(rc.getColor());



		    int x1 = xpos + CB_WD + GAP;

		    int y1 = ypos+HEIGHT/2;

		    int x2 = x1+WIDTH;

		    int y2 = y1;



		    Rectangle rect = 

			new Rectangle(r.x + xpos + CB_WD + GAP,

				      r.y + ypos, WIDTH, HEIGHT);

		 

		    relnClass[n] = rc; 

		    relnChk[n].resize(CB_WD, CB_HT); 

		    relnChk[n].move(r.x + GAP, r.y + 

					tabDim.height + ypos + GAP + 3);

		    relnChk[n].show();

		    relnChk[n].setState(rc.isVisible());



		    objects.addElement(new LegendObject(rect, 

					rc.getDescription()));



		    RealPoint p1 = new RealPoint(x1, y1); 

		    RealPoint p2 = new RealPoint(x2, y2); 



		    Util.drawArrow(g, p1, p2);



		    g.setColor(Color.black);

		    Util.drawStringClipped(g, str, x1+WIDTH+GAP, ypos, 

						tw, h);

                    ypos += h+GAP;

		}

		item++;

		n++;

	    }

        }



	for (int i=n; i<MAX_NUM_CHECKS; i++) {

	    relnChk[i].hide();

	    relnClass[i] = null;

	}



	if (n == 0) {

	    g.setColor(Color.black);



	    int x1 = xpos;

	    int y1 = ypos+HEIGHT/2;

	    int x2 = xpos+WIDTH;

	    int y2 = ypos+HEIGHT/2;



	    RealPoint p1 = new RealPoint(x1, y1);

	    RealPoint p2 = new RealPoint(x2, y2);



	    Util.drawArrow(g, p1, p2);



	    g.setColor(Color.black);

	    Util.drawStringClipped(g, "Relation", xpos+WIDTH+GAP, ypos, tw, h);



	    ypos += h+GAP;



	    item++;

	}



	if (fullInfo) {

	    ypos += h/2;



	    Util.drawStringClipped(g, endRelMsg1, xpos, ypos, tw, h);



	    ypos += h;



	    Util.drawStringClipped(g, endRelMsg2, xpos, ypos, tw, h);

	}



	setupScrolling(ypos-iypos);

    }





    protected void genChecks() {

	for (int i=0; i<MAX_NUM_CHECKS; i++) {

	    relnChk[i] = new Checkbox();

	    relnChk[i].hide();

	    ls.add(relnChk[i]);

	}

    }





    // --------------

    // Public methods 

    // --------------



    public LegendBox(LandscapeViewerCore ls) {

	super(ls, "Legend", 5);

	genChecks();

    }



    public boolean mouseMotion(Event ev, int x, int y) {

	Enumeration en = objects.elements();



	while (en.hasMoreElements()) {

	    LegendObject lo = (LegendObject) en.nextElement();



	    if (x >= lo.r.x && x <= lo.r.x + lo.r.width &&

		y >= lo.r.y && y <= lo.r.y + lo.r.height)

	    {

		if (lo.description != null) {

		    ls.setHelpPopup(lo.description, x, y);

		    return true;

		}

	    }

	}



	// Otherwise, display the general help



	ls.setHelpPopup(helpStr, x, y);

	return true;

    }



    public boolean handleEvent(Diagram dg, Graphics gc, Event event) {

	if (!isVisible())

	    return false;



	if (super.handleEvent(dg, gc, event)) {

	    return true;

	}



	for(int i = 0; i<MAX_NUM_CHECKS; i++) {

	    if (event.target == relnChk[i]) {

		if (event.id == Event.ACTION_EVENT) {

		    ls.processKey('1' + i, 0);

		}



		return true;

	    }

	}

	return false;

    }



    public void toggleRelationVisibility(Diagram dg, int key) {

	if (key == 10) {

	    // Affect all relations



	    boolean ns = !relnClass[0].isVisible();



	    int i = 0;



	    while(relnClass[i] != null) {

		dg.setEdgeVisibilityState(relnClass[i], ns);

		i++;

	    }

	}

	else {

	    RelationClass rc = relnClass[key];



	    if (rc != null) {

	        boolean ns = !rc.isVisible();



	        dg.setEdgeVisibilityState(rc, ns);

	        ls.doFeedback("Relation " + rc.getLabel()

				+ " set to " + (ns ? "visible" : "hidden"));

	    }

	}

	repaint();

    }





    public void activate() {

	super.activate();

    }



    public void inactivate() {

	super.inactivate();



	for (int i=0; i<MAX_NUM_CHECKS; i++) {

	    relnChk[i].hide();

	}

    }

}





