package lsedit;

import java.awt.*;

import java.util.*;



public class ResizeModeHandler extends LandscapeModeHandler

{

    // Variables used by modes



    protected static final int RM_NONE   = 0;

    protected static final int RM_LEFT   = 1;

    protected static final int RM_RIGHT  = 2;

    protected static final int RM_TOP    = 4;

    protected static final int RM_BOTTOM = 8;



    protected static final int LR_MASK = RM_LEFT | RM_RIGHT;

    protected static final int TB_MASK = RM_TOP | RM_BOTTOM;



    protected static final int RZONE_SIZE = 5;



    protected static final double MIN_DIM    = 10.0;



    protected Layout curLayout;

    protected Layout bb; 		// bounding box of children 



    protected int resizeMode = RM_NONE;



    protected int[] startMode = 

	{ RM_TOP | RM_LEFT, RM_TOP, RM_TOP | RM_RIGHT, RM_RIGHT,

	  RM_BOTTOM | RM_RIGHT, RM_BOTTOM, RM_BOTTOM | RM_LEFT, RM_LEFT

	};



    protected void addMode(int newMode) {

	int omode = 0;



	switch(newMode) 

	{

	case RM_TOP:

	    omode = RM_BOTTOM;

	    break;



	case RM_BOTTOM:

	    omode = RM_TOP;

	    break;



	case RM_LEFT:

	    omode = RM_RIGHT;

	    break;



	case RM_RIGHT:

	    omode = RM_LEFT;

	    break;

        } 



	if ((resizeMode & omode) == 0) {

	    resizeMode |= newMode;

	}

    }



    protected void doResizeAdjust(int mode, double xpos, double ypos) {



	// Sanitize to parent



	Layout plyt = e.getParent().getLayout();



	if (xpos < plyt.x + LandscapeViewerCore.GAP) {

	    xpos = plyt.x + LandscapeViewerCore.GAP;

	}

	else if (xpos > plyt.x + plyt.width - LandscapeViewerCore.GAP)

	{

	    xpos = plyt.x + plyt.width - LandscapeViewerCore.GAP;

	}



	if (ypos < plyt.y + LandscapeViewerCore.GAP) {

	    ypos = plyt.y + LandscapeViewerCore.GAP;

	}

	else if (ypos > plyt.y + plyt.height - LandscapeViewerCore.GAP)

	{

	    ypos = plyt.y + plyt.height - LandscapeViewerCore.GAP;

	}



        switch(mode)

        {

        case RM_TOP:

            if (curLayout.height + curLayout.y - ypos >= MIN_DIM) {

                curLayout.height += (curLayout.y-ypos);

                curLayout.y = ypos;

            }

            if (xpos-curLayout.x < 4) 

		addMode(RM_LEFT);

            else if (curLayout.x + curLayout.width - xpos < 4)

		addMode(RM_RIGHT);

            break;



        case RM_BOTTOM:

            if (ypos - curLayout.y >= MIN_DIM) {

                curLayout.height = (ypos-curLayout.y);

            }

            if (xpos-curLayout.x < 4)

                addMode(RM_LEFT);

            else if (curLayout.x + curLayout.width - xpos < 4)

                addMode(RM_RIGHT);

            break;



        case RM_LEFT:

            if (curLayout.width + curLayout.x - xpos >= MIN_DIM) {

                curLayout.width += (curLayout.x-xpos);

                curLayout.x = xpos;

            }

            if (ypos-curLayout.y < 4)

                addMode(RM_TOP);

            else if (curLayout.y + curLayout.height - ypos < 4)

                addMode(RM_BOTTOM);

            break;



        case RM_RIGHT:

            if (xpos - curLayout.x >= MIN_DIM) {

                curLayout.width = (xpos-curLayout.x);

            }

            if (ypos-curLayout.y < 4)

                addMode(RM_TOP);

            else if (curLayout.y + curLayout.height - ypos < 4)

                addMode(RM_BOTTOM);

            break;

        }

    }



    //

    // Public methods

    //



    public ResizeModeHandler() {

    }



    //

    //   ++-----------++

    //   +             +   

    //   |             |  Corner or edge grab

    //   |             | 

    //   +             +

    //   ++-----------++

    //



    public boolean mouseDown(Event ev, int x, int y) {

	e = dg.mouseOver(x, y);



        if (e == null || !e.getGroupKeyFlag()) {

            return false;

        } 



	int rpoint = e.overResizeTab(x, y);



	if (rpoint < 0)

	    return false;



	resizeMode = startMode[rpoint];



	curLayout = (Layout) e.getLayout().clone();



        gc.setXORMode(e.getParent().getBackground());

        dg.drawEntityOutline(gc, curLayout, e);



	ls.doFeedback("Size: " + Math.round(curLayout.width) + " x " +

				 Math.round(curLayout.height));

	return true; 

    }



    public boolean mouseDrag(Event ev, int x, int y) {

        double xpos = x;

        double ypos = y;



        if (resizeMode == RM_NONE) {

            // See if we have reached an edge



            boolean l = (xpos <= curLayout.x);

            boolean r = (xpos >= curLayout.x + curLayout.width);

            boolean t = (ypos <= curLayout.y);

            boolean b = (ypos >= curLayout.y + curLayout.height);



            if (e == dg.getRoot()) {

                l = false;

                t = false;

            }



            if (l) {

                if (t || ypos-curLayout.y < 4)

                    resizeMode = RM_LEFT | RM_TOP;

                else if (b || (curLayout.y + curLayout.height - ypos < 4))

                    resizeMode = RM_LEFT | RM_BOTTOM;

                else

                    resizeMode = RM_LEFT;

            }

            else if (r) {

                if (t || ypos-curLayout.y < 4)

                    resizeMode = RM_RIGHT | RM_TOP;

                else if (b || (curLayout.y + curLayout.height - ypos < 4))

                    resizeMode = RM_RIGHT | RM_BOTTOM;

                else

                    resizeMode = RM_RIGHT;

            }

            else if (t) {

                if (xpos-curLayout.x < 4)

                    resizeMode = RM_TOP | RM_LEFT;

                else if (curLayout.x + curLayout.width - xpos < 4)

                    resizeMode = RM_TOP | RM_RIGHT;

                else

                    resizeMode = RM_TOP;

            }

            else if (b) {

                if (xpos-curLayout.x < 4)

                    resizeMode = RM_BOTTOM | RM_LEFT;

                else if (curLayout.x + curLayout.width - xpos < 4)

                    resizeMode = RM_BOTTOM | RM_RIGHT;

                else

                    resizeMode = RM_BOTTOM;

            }

        }



        if (resizeMode == RM_NONE)

            return true; 



        dg.drawEntityOutline(gc, curLayout, e);



        doResizeAdjust((resizeMode & TB_MASK), xpos, ypos);

        doResizeAdjust((resizeMode & LR_MASK), xpos, ypos);



        dg.drawEntityOutline(gc, curLayout, e);



	ls.doFeedback("Size: " + Math.round(curLayout.width) + " x " +

				 Math.round(curLayout.height));

	return true;

    }



    public void mouseUp(Event ev, int x, int y) {

	if (e != null) {

            dg.saveForUndo();

            e.setLayout(curLayout);

	    dg.rescale();

	    e.fitTo(false);

            ls.redrawDg();

	}



	ls.clearFeedback();

    }

}

